// ------------------------------
// Filename      : landscape.cpp
// Author        : Sven Maerivoet
// Last modified : 30/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifdef _WIN32
  #include <windows.h>
#endif

#include "landscape.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

using namespace std;

// --------------------
// class LandscapeGLApp
// --------------------

// -----------
// constructor
// -----------
LandscapeGLApp::LandscapeGLApp(void) : fSeaTexture(0),
                                       fMountainTexture(0),
                                       fSnowTexture(0)
{
}

// ----------
// destructor
// ----------
LandscapeGLApp::~LandscapeGLApp(void)
{
  delete fSeaTexture;
  delete fMountainTexture;
  delete fSnowTexture;
}

// -----------------------
// explicit initialisation
// -----------------------
void LandscapeGLApp::explicitInitialize(void)
{
  fFogEnabled = false;

  fUseLandscapeMaterials = true;
  fUseTextureMaps = false;
  fSmoothenVertexNormals = false;

  fSeaTexture = new RGBApixmap();
  if (!fSeaTexture->loadFromBMPFile("data/textures/other/sea.bmp")) {
    exit(0);
  }
  fSeaTexture->setTextureID(getUniqueTextureID());

  fMountainTexture = new RGBApixmap();
  if (!fMountainTexture->loadFromBMPFile("data/textures/other/mountain.bmp")) {
    exit(0);
  }
  fMountainTexture->setTextureID(getUniqueTextureID());

  fSnowTexture = new RGBApixmap();
  if (!fSnowTexture->loadFromBMPFile("data/textures/other/snow.bmp")) {
    exit(0);
  }
  fSnowTexture->setTextureID(getUniqueTextureID());

  fTimePeriod = 20;
  fCurrentTime = 0;

  fBounds.set(-100.0,-100.0,100.0,100.0);
  fTerrainMap.createFractalTerrain(16,
                                   10.0,10.0,
                                   10.0,10.0,
                                   150.0,
                                   0.6,
                                   fBounds);

  fTerrainMap.setSeaLevel(0.0);
  fTerrainMap.setSnowLevel(50.0);
  fTerrainMap.setTextures(fSeaTexture,fMountainTexture,fSnowTexture);
  fTerrainMap.setCurrentTime(fCurrentTime);

  setCameraSlideStepSize(5.0);
  setCameraRotateStepSize(1.0);

  setCamera(Point3D(150.0,350.0,150.0),
            Point3D(0.0,0.0,0.0),
            Vector3D(0.0,1.0,0.0));
  setCameraPerspectiveShape(45.0,getWorldWindowAspectRatio(),
                            1.0,10000.0);
}

// ------------------
// callback-functions
// ------------------
void LandscapeGLApp::displayFunc(void)
{
  if (!fFogEnabled) {
    // clear the screen with a deep sort of blue color
    setBackgroundColor(Color(0.2,0.0,0.3));
  }
  else {
    // clear the screen with the gray color of the fog
    Color fogColor(0.4,0.4,0.4);
    setBackgroundColor(fogColor);
    enableExponentialFog(fogColor,0.005);
  }
  clearScreen();

  switchLightOn(kLight1);
  setLightDirection(kLight1,Vector3D(getCameraEyePoint().fX,
                                     getCameraEyePoint().fY,
                                     getCameraEyePoint().fZ));
  setLightProperties(kLight1,
                     Color(0.0,0.0,0.0),
                     Color(1.0,1.0,1.0),
                     Color(1.0,1.0,1.0));

  if (fUseTextureMaps) {
    setMaterial(Color(1.0,1.0,1.0),
                Color(1.0,1.0,1.0),
                Color(1.0,1.0,1.0),
                128.0);
    drawLandscape(fTerrainMap,
                  fUseLandscapeMaterials,
                  fSmoothenVertexNormals,
                  kApplyTextureMapping);
  }
  else {
    setMaterialToGold();
    drawLandscape(fTerrainMap,
                  fUseLandscapeMaterials,
                  fSmoothenVertexNormals,
                  kDontApplyTextureMapping);
  }
}

void LandscapeGLApp::keyboardFunc(unsigned char key, int x, int y)
{
  TOpenGLApp::keyboardFunc(key,x,y);

  switch (key) {
    case 'j' :
      if (fFogEnabled) {
        fFogEnabled = false;
        disableFog();
      }
      break;

    case 'J' :
      if (!fFogEnabled) {
        fFogEnabled = true;
      }
      break;

    case 'o' :
      fTerrainMap.setSnowLevel(fTerrainMap.getSnowLevel() - 2.0);
      break;

    case 'O' :
      fTerrainMap.setSnowLevel(fTerrainMap.getSnowLevel() + 2.0);
      break;

    case '-' :
      fTerrainMap.setSeaLevel(fTerrainMap.getSeaLevel() - 2.0);
      break;

    case '+' :
      fTerrainMap.setSeaLevel(fTerrainMap.getSeaLevel() + 2.0);
      break;

    case 'd' :
    case 'D' :
      fTerrainMap.disableSea();
      break;

    case '1' :
      fTerrainMap.setWaveHeight(fTerrainMap.getWaveHeight() - 1.0);
      break;

    case '7' :
      fTerrainMap.setWaveHeight(fTerrainMap.getWaveHeight() + 1.0);
      break;

    case '2' :
      if (fTerrainMap.getWavePeriod() > 1) {
        fTerrainMap.setWavePeriod(fTerrainMap.getWavePeriod() - 1);
        --fTimePeriod;
      }
      break;

    case '8' :
      fTerrainMap.setWavePeriod(fTerrainMap.getWavePeriod() + 1);
      ++fTimePeriod;
      ;

    case '3' :
      if (fTerrainMap.getSeaFrequency() > 1) {
        fTerrainMap.setSeaFrequency(fTerrainMap.getSeaFrequency() - 1);
      }
      break;

    case '9' :
      fTerrainMap.setSeaFrequency(fTerrainMap.getSeaFrequency() + 1);
      break;

    case 'i' :
      fTerrainMap.createFractalTerrain(fTerrainMap.getUSize() / 2,
                                       10.0,10.0,
                                       10.0,10.0,
                                       150.0,
                                       0.6,
                                       fBounds);
      break;

    case 'I' :
      fTerrainMap.createFractalTerrain(fTerrainMap.getUSize() * 2,
                                       10.0,10.0,
                                       10.0,10.0,
                                       150.0,
                                       0.6,
                                       fBounds);
      break;

    case '4' :
      fTerrainMap.createFractalTerrain(16,
                                       10.0,10.0,
                                       10.0,10.0,
                                       150.0,
                                       0.6,
                                       fBounds);
      break;

    case '5' :
      fTerrainMap.constructFromBMPFile("data/textures/plasma/plasma2.bmp",100.0,fBounds);
      break;

    case '6' :
      fTerrainMap.constructFromBMPFile("data/textures/other/mandelbrot-filled.bmp",100.0,fBounds);
      break;

    case 'r' :
      fTerrainMap.useSpecifiedColors();
      break;

    case 'R' :
      fTerrainMap.useRealColors();
      break;

    case 'w' :
      fUseLandscapeMaterials = false;
      break;

    case 'W' :
      fUseLandscapeMaterials = true;
      break;

    case 'b' :
      fUseTextureMaps = false;
      break;

    case 'B' :
      fUseTextureMaps = true;
      break;

    case 'g' :
      disableGlowingTextures();
      break;

    case 'G' :
      enableGlowingTextures();
      break;

    case 'u' :
      if (fSmoothenVertexNormals) {
        fSmoothenVertexNormals = false;
      }
      break;

    case 'U' :
      if (!fSmoothenVertexNormals) {
        fSmoothenVertexNormals = true;
      }
      break;

    case 'p' :
    case 'P' :
      // write mesh to 3VN-file landscape.3vn
      fTerrainMap.writeTo3VNMeshFile("landscape.3vn");
      break;
  }

  if (fScheduleRedisplay) {
    redisplay();
  }
}

void LandscapeGLApp::idleFunc(void)
{
  TOpenGLApp::idleFunc();

  fTerrainMap.setCurrentTime(fCurrentTime);
  ++fCurrentTime;
  if (fCurrentTime >= fTimePeriod) {
    fCurrentTime = 0;
  }

  redisplay();
}
