// ------------------------------
// Filename      : affine.cpp
// Author        : Sven Maerivoet
// Last modified : 30/04/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#include "affine.h"
#include "math.h"
#include <math.h>

using namespace std;

// --------------------------
// class AffineTransformation
// --------------------------

// ------------
// constructors
// ------------
AffineTransformation::AffineTransformation(void)
{
  setIdentityMatrix();
}

AffineTransformation::AffineTransformation(const AffineTransformation& affineTransformation)
{
  for (int i = 0; i < 16; ++i) {
    m[i] = affineTransformation.m[i];
  }
}

// ---------
// destructor
// ----------
AffineTransformation::~AffineTransformation(void)
{
}

// -------
// methods
// -------
AffineTransformation& AffineTransformation::operator=(const AffineTransformation& affineTransformation)
{
  // don't assign an affine transformation to itself
  if (this != &affineTransformation) {
    set(affineTransformation);
  }

  return *this;
}

void AffineTransformation::set(const AffineTransformation& affineTransformation)
{
  for (int i = 0; i < 16; ++i) {
    m[i] = affineTransformation.m[i];
  }
}

void AffineTransformation::setIdentityMatrix(void)
{
  m[0] = 1.0; m[4] = 0.0; m[8]  = 0.0; m[12] = 0.0;
  m[1] = 0.0; m[5] = 1.0; m[9]  = 0.0; m[13] = 0.0;
  m[2] = 0.0; m[6] = 0.0; m[10] = 1.0; m[14] = 0.0;
  m[3] = 0.0; m[7] = 0.0; m[11] = 0.0; m[15] = 1.0;
}

void AffineTransformation::preMult(const AffineTransformation& affineTransformation)
{
  AffineTransformation temporary; 
  temporary.set(*this);

  for (int col = 0; col < 4; ++col) {
    for (int row = 0; row < 4; ++row) {

      double sum = 0.0;
      for (int k = 0; k < 4; ++k) {
        sum += affineTransformation.m[(4 * k) + row] * temporary.m[(4 * col) + k];
        m[(4 * col) + row] = sum;
      }
    }
  }
}

void AffineTransformation::postMult(const AffineTransformation& affineTransformation)
{
  AffineTransformation temporary; 
  temporary.set(*this);

  for (int col = 0; col < 4; ++col) {
    for (int row = 0; row < 4; ++row) {

      double sum = 0.0;
      for (int k = 0; k < 4; ++k) {
        sum += temporary.m[(4 * k) + row] * affineTransformation.m[(4 * col) + k];
        m[(4 * col) + row] = sum;
      }
    }
  }
}

AffineTransformation AffineTransformation::operator*(double scalar) const
{
  AffineTransformation affineTransformation(*this);
  for (int i = 0; i < 16; ++i) {
    affineTransformation.m[i] *= scalar;
  }

  return affineTransformation;
}

AffineTransformation& AffineTransformation::operator*=(double scalar)
{
  for (int i = 0; i < 16; ++i) {
    m[i] *= scalar;
  }

  return *this;
}

AffineTransformation AffineTransformation::convertRotationToTransformation(double angleInDegrees, const Vector3D& u)
{
  AffineTransformation affineTransformation;

  Vector3D un(u);
  un.normalize();
  double rad = Math::degToRad(angleInDegrees);
  double c = cos(rad);
  double s = sin(rad);
  double mc = 1.0 - c;

  affineTransformation.m[0]  = c + (mc * un.fX * un.fX);
  affineTransformation.m[1]  = (mc * un.fX * un.fY) + (s * un.fZ);
  affineTransformation.m[2]  = (mc * un.fX * un.fZ) - (s * un.fY);
  affineTransformation.m[4]  = (mc * un.fY * un.fX) - (s * un.fZ);
  affineTransformation.m[5]  = c + (mc * un.fY * un.fY);
  affineTransformation.m[6]  = (mc * un.fY * un.fZ) + (s * un.fX);
  affineTransformation.m[8]  = (mc * un.fZ * un.fX) + (s * un.fY);
  affineTransformation.m[9]  = (mc * un.fZ * un.fY) - (s * un.fX);
  affineTransformation.m[10] = c + (mc * un.fZ * un.fZ);

  return affineTransformation;
}

AffineTransformation AffineTransformation::convertScalingToTransformation(double sx, double sy, double sz)
{
  AffineTransformation affineTransformation;

  affineTransformation.m[0]  = sx; 
  affineTransformation.m[5]  = sy; 
  affineTransformation.m[10] = sz;

  return affineTransformation;
}

AffineTransformation AffineTransformation::convertTranslationToTransformation(const Vector3D& direction)
{
  AffineTransformation affineTransformation;

  affineTransformation.m[12] = direction.fX;
  affineTransformation.m[13] = direction.fY;
  affineTransformation.m[14] = direction.fZ;

  return affineTransformation;
}

// ------------------------------
// class AffineTransformationNode
// ------------------------------

// -----------
// constructor
// -----------
AffineTransformationNode::AffineTransformationNode(void) : fNext(0)
{
  fAffineTransformation = new AffineTransformation();
  fInverseAffineTransformation = new AffineTransformation();
}

// ---------
// destructor
// ----------
AffineTransformationNode::~AffineTransformationNode(void)
{
  delete fAffineTransformation;
  delete fInverseAffineTransformation;
}

// -------------------------------
// class AffineTransformationStack
// -------------------------------

// -----------
// constructor
// -----------
AffineTransformationStack::AffineTransformationStack(void)
{
  // push the identity-transformation on the stack
  fTopOfStack = new AffineTransformationNode();
}

// ---------
// destructor
// ----------
AffineTransformationStack::~AffineTransformationStack(void)
{
  clear();
}

// -------
// methods
// -------
void AffineTransformationStack::duplicate(void)
{
  AffineTransformationNode* temporary = new AffineTransformationNode();

  temporary->fAffineTransformation = new AffineTransformation(*(fTopOfStack->fAffineTransformation));
  temporary->fInverseAffineTransformation = new AffineTransformation(*(fTopOfStack->fInverseAffineTransformation));
  temporary->fNext = fTopOfStack;
  fTopOfStack = temporary;
}

void AffineTransformationStack::setIdentity(void)
{
  if (fTopOfStack != 0) {
    fTopOfStack->fAffineTransformation->setIdentityMatrix();
    fTopOfStack->fInverseAffineTransformation->setIdentityMatrix();
  }
}

void AffineTransformationStack::pop(void)
{
  if (fTopOfStack != 0) {
    AffineTransformationNode* temporary = fTopOfStack;
    fTopOfStack = fTopOfStack->fNext;
    delete temporary;
  }
}

void AffineTransformationStack::clear(void)
{
  while (fTopOfStack != 0) {
    pop();
  }
}

void AffineTransformationStack::rotate(double angleInDegrees, const Vector3D& u)
{
  AffineTransformation
    rotation(AffineTransformation::convertRotationToTransformation(angleInDegrees,u));
  AffineTransformation
    inverseRotation(AffineTransformation::convertRotationToTransformation(-angleInDegrees,u));

  fTopOfStack->fAffineTransformation->postMult(rotation);
  fTopOfStack->fInverseAffineTransformation->preMult(inverseRotation);
}

void AffineTransformationStack::scale(double sx, double sy, double sz)
{
  AffineTransformation
    scaler(AffineTransformation::convertScalingToTransformation(sx,sy,sz));

  if ((Math::abs(sx) < Math::kEpsilon) ||
      (Math::abs(sy) < Math::kEpsilon) ||
      (Math::abs(sz) < Math::kEpsilon)) {
    // degenerate scaling transformation
    return;
  }

  AffineTransformation
    inverseScaler(AffineTransformation::convertScalingToTransformation(1.0 / sx,
                                                                       1.0 / sy,
                                                                       1.0 / sz));

  fTopOfStack->fAffineTransformation->postMult(scaler);
  fTopOfStack->fInverseAffineTransformation->preMult(inverseScaler);
}
		
void AffineTransformationStack::translate(const Vector3D& direction)
{
  AffineTransformation
    translation(AffineTransformation::convertTranslationToTransformation(direction));

  AffineTransformation
    inverseTranslation(AffineTransformation::convertTranslationToTransformation(direction * (-1.0)));

  fTopOfStack->fAffineTransformation->postMult(translation);
  fTopOfStack->fInverseAffineTransformation->preMult(inverseTranslation);
}

void AffineTransformationStack::transform(const Vector3D& row1,
                                          const Vector3D& row2,
                                          const Vector3D& row3)
{
  AffineTransformation transformation;
  AffineTransformation inverseTransformation;

  transformation.m[0]  = row1.fX;
  transformation.m[4]  = row1.fY;
  transformation.m[8]  = row1.fZ;
  transformation.m[12] = 0.0;

  transformation.m[1]  = row2.fX;
  transformation.m[5]  = row2.fY;
  transformation.m[9]  = row2.fZ;
  transformation.m[13] = 0.0;

  transformation.m[2]  = row3.fX;
  transformation.m[6]  = row3.fY;
  transformation.m[10] = row3.fZ;
  transformation.m[14] = 0.0;

  transformation.m[3]  = 0.0;
  transformation.m[7]  = 0.0;
  transformation.m[11] = 0.0;
  transformation.m[15] = 1.0;

  // calculate inverse transformation
  double determinant = (transformation.m[0] * transformation.m[5] * transformation.m[10]) +
                       (transformation.m[1] * transformation.m[6] * transformation.m[8]) +
                       (transformation.m[4] * transformation.m[9] * transformation.m[2]) -
                       (transformation.m[2] * transformation.m[5] * transformation.m[8]) -
                       (transformation.m[6] * transformation.m[9] * transformation.m[0]) -
                       (transformation.m[1] * transformation.m[4] * transformation.m[10]);

  if (Math::abs(determinant) < Math::kEpsilon) {
    // singular matrix
    return;
  }

  // fill in the transposed cofactors
  inverseTransformation.m[0] = (transformation.m[5] * transformation.m[10]) -
                               (transformation.m[6] * transformation.m[9]);
  inverseTransformation.m[1] = (transformation.m[2] * transformation.m[9]) -
                               (transformation.m[1] * transformation.m[0]);
  inverseTransformation.m[2] = (transformation.m[1] * transformation.m[6]) -
                               (transformation.m[2] * transformation.m[5]);
  inverseTransformation.m[3] = 0.0;

  inverseTransformation.m[4] = (transformation.m[6] * transformation.m[8]) -
                               (transformation.m[4] * transformation.m[10]);
  inverseTransformation.m[5] = (transformation.m[0] * transformation.m[10]) -
                               (transformation.m[2] * transformation.m[8]);
  inverseTransformation.m[6] = (transformation.m[2] * transformation.m[4]) -
                               (transformation.m[0] * transformation.m[6]);
  inverseTransformation.m[7] = 0.0;

  inverseTransformation.m[8] = (transformation.m[4] * transformation.m[9]) -
                               (transformation.m[5] * transformation.m[8]);
  inverseTransformation.m[9] = (transformation.m[1] * transformation.m[8]) -
                               (transformation.m[0] * transformation.m[9]);
  inverseTransformation.m[10] = (transformation.m[0] * transformation.m[5]) -
                                (transformation.m[1] * transformation.m[4]);
  inverseTransformation.m[11] = 0.0;

  inverseTransformation.m[12] = 0.0;
  inverseTransformation.m[13] = 0.0;
  inverseTransformation.m[14] = 0.0;
  inverseTransformation.m[15] = determinant;

  // divide each component by the determinant
  for (int i = 0; i < 16; ++i) {
    inverseTransformation.m[i] /= determinant;
  }

  fTopOfStack->fAffineTransformation->postMult(transformation);
  fTopOfStack->fInverseAffineTransformation->preMult(inverseTransformation);
}
