// ------------------------------
// Filename      : extents.cpp
// Author        : Sven Maerivoet
// Last modified : 25/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#include "extents.h"
#include "math.h"
#include <math.h>

using namespace std;

// ------------------
// class PointCluster
// ------------------

// ------------
// constructors
// ------------
PointCluster::PointCluster(void) : fNrOfPoints(0)
{
  fPoints = 0;
}

PointCluster::PointCluster(int nrOfPoints) : fNrOfPoints(nrOfPoints)
{
  fPoints = new Point3D[fNrOfPoints];
}

// ----------
// destructor
// ----------
PointCluster::~PointCluster(void)
{
  delete[] fPoints;
  fPoints = 0;
}

// -------
// methods
// -------
Point3D PointCluster::calcCentroid(void) const
{
  Point3D centroid(0.0,0.0,0.0);

  for (int i = 0; i < fNrOfPoints; ++i) {
    centroid.fX += fPoints[i].fX;
    centroid.fY += fPoints[i].fY;
    centroid.fZ += fPoints[i].fZ;
  }

  centroid.fX /= (double) fNrOfPoints;
  centroid.fY /= (double) fNrOfPoints;
  centroid.fZ /= (double) fNrOfPoints;

  return centroid;
}

void PointCluster::applyAffineTransformation(const AffineTransformation& affine)
{
  for (int i = 0; i < fNrOfPoints; ++i) {
    Point3D original(fPoints[i]);
    Transformations::transformPoint(fPoints[i],affine,original);
  }
}

void PointCluster::applyTranslation(const AffineTransformation& affine)
{
  for (int i = 0; i < fNrOfPoints; ++i) {
    Point3D original(fPoints[i]);
    Transformations::translatePoint(fPoints[i],affine,original);
  }
}

// -------------------
// class GenericExtent
// -------------------

// -----------
// constructor
// -----------
GenericExtent::GenericExtent(void)
{
}

// ----------
// destructor
// ----------
GenericExtent::~GenericExtent(void)
{
}

// -------
// methods
// -------
void GenericExtent::create(const PointCluster& pointCluster, double excess)
{
}

bool GenericExtent::rayIntersectsExtent(const Ray& ray) const
{
  return false;
}

void GenericExtent::draw(void) const
{
}

// ---------------
// class BoxExtent
// ---------------

// -----------
// constructor
// -----------
BoxExtent::BoxExtent(void)
{
}

// ----------
// destructor
// ----------
BoxExtent::~BoxExtent(void)
{
}

// -------
// methods
// -------
void BoxExtent::create(const PointCluster& pointCluster, double excess)
{
  fLeft = pointCluster.fPoints[0].fX;
  fRight = pointCluster.fPoints[0].fX;
  fBottom = pointCluster.fPoints[0].fY;
  fTop = pointCluster.fPoints[0].fY;
  fBack = pointCluster.fPoints[0].fZ;
  fFront = pointCluster.fPoints[0].fZ;

  // determine the largest extent around the pointcluster
  for (int i = 1; i < pointCluster.fNrOfPoints; ++i) {

    if (pointCluster.fPoints[i].fX < fLeft) {
      fLeft = pointCluster.fPoints[i].fX;
    }
    else if (pointCluster.fPoints[i].fX > fRight) {
      fRight = pointCluster.fPoints[i].fX;
    }

    if (pointCluster.fPoints[i].fY < fBottom) {
      fBottom = pointCluster.fPoints[i].fY;
    }
    else if (pointCluster.fPoints[i].fY > fTop) {
      fTop = pointCluster.fPoints[i].fY;
    }

    if (pointCluster.fPoints[i].fZ < fBack) {
      fBack = pointCluster.fPoints[i].fZ;
    }
    else if (pointCluster.fPoints[i].fZ > fFront) {
      fFront = pointCluster.fPoints[i].fZ;
    }
  }

  // apply the excess
  fLeft -= (excess / 2.0);
  fRight += (excess / 2.0);
  fBottom -= (excess / 2.0);
  fTop += (excess / 2.0);
  fBack -= (excess / 2.0);
  fFront += (excess / 2.0);
}

bool BoxExtent::rayIntersectsExtent(const Ray& ray) const
{
  double tNear = -Math::kInfinity;
  double tFar = Math::kInfinity;

  if (Math::abs(ray.fDirection.fX) <= Math::kEpsilon) { // the ray's direction is parallel to the X-planes
    if ((ray.fStart.fX < fLeft) ||
        (ray.fStart.fX > fRight)) { // the ray's origin doesn't lie between the X-planes
      return false;
    }
  }
  else { // the ray's direction is not parallel to the X-planes
    double t1 = (fLeft - ray.fStart.fX) / ray.fDirection.fX;
    double t2 = (fRight - ray.fStart.fX) / ray.fDirection.fX;
    if (t1 > t2) {
      double temp = t1;
      t1 = t2;
      t2 = temp;
    }
    
    if (t1 > tNear) {
      tNear = t1;
    }
    if (t2 < tFar) {
      tFar = t2;
    }

    if (tNear > tFar) {
      return false; // box is missed
    }
    if (tFar < 0.0) {
      return false; // box is behind the ray
    }
  }

  if (Math::abs(ray.fDirection.fY) <= Math::kEpsilon) { // the ray's direction is parallel to the Y-planes
    if ((ray.fStart.fY < fBottom) ||
        (ray.fStart.fY > fTop)) { // the ray's origin doesn't lie between the Y-planes
      return false;
    }
  }
  else { // the ray's direction is not parallel to the Y-planes
    double t1 = (fBottom - ray.fStart.fY) / ray.fDirection.fY;
    double t2 = (fTop - ray.fStart.fY) / ray.fDirection.fY;
    if (t1 > t2) {
      double temp = t1;
      t1 = t2;
      t2 = temp;
    }
    
    if (t1 > tNear) {
      tNear = t1;
    }
    if (t2 < tFar) {
      tFar = t2;
    }

    if (tNear > tFar) {
      return false; // box is missed
    }
    if (tFar < 0.0) {
      return false; // box is behind the ray
    }
  }

  if (Math::abs(ray.fDirection.fZ) <= Math::kEpsilon) { // the ray's direction is parallel to the Z-planes
    if ((ray.fStart.fZ < fBack) ||
        (ray.fStart.fZ > fFront)) { // the ray's origin doesn't lie between the Z-planes
      return false;
    }
  }
  else { // the ray's direction is not parallel to the Z-planes
    double t1 = (fBack - ray.fStart.fZ) / ray.fDirection.fZ;
    double t2 = (fFront - ray.fStart.fZ) / ray.fDirection.fZ;
    if (t1 > t2) {
      double temp = t1;
      t1 = t2;
      t2 = temp;
    }
    
    if (t1 > tNear) {
      tNear = t1;
    }
    if (t2 < tFar) {
      tFar = t2;
    }

    if (tNear > tFar) {
      return false; // box is missed
    }
    if (tFar < 0.0) {
      return false; // box is behind the ray
    }
  }

  return true;  
}

void BoxExtent::draw(void) const
{
  glPushMatrix();
    glTranslated((GLdouble) ((fLeft + fRight) / 2.0),
                 (GLdouble) ((fBottom + fTop) / 2.0),
                 (GLdouble) ((fBack + fFront) / 2.0));
    glScaled((GLdouble) Math::abs(fRight - fLeft),
             (GLdouble) Math::abs(fTop - fBottom),
             (GLdouble) Math::abs(fFront - fBack));
    glutWireCube((GLdouble) 1.0);
  glPopMatrix();
}

// ------------------
// class SphereExtent
// ------------------

// -----------
// constructor
// -----------
SphereExtent::SphereExtent(void)
{
}

// ----------
// destructor
// ----------
SphereExtent::~SphereExtent(void)
{
}

// -------
// methods
// -------
void SphereExtent::create(const PointCluster& pointCluster, double excess)
{
  fOrigin = pointCluster.calcCentroid();
  fRadius = 0.0;

  // determine the largest distance from the centroid to any of the points in the pointcluster
  for (int i = 0; i < pointCluster.fNrOfPoints; ++i) {

    double distance = sqrt(((fOrigin.fX - pointCluster.fPoints[i].fX) *
                            (fOrigin.fX - pointCluster.fPoints[i].fX)) + 
                           ((fOrigin.fY - pointCluster.fPoints[i].fY) *
                            (fOrigin.fY - pointCluster.fPoints[i].fY)) +
                           ((fOrigin.fZ - pointCluster.fPoints[i].fZ) *
                            (fOrigin.fZ - pointCluster.fPoints[i].fZ)));

    if (distance > fRadius) {
      fRadius = distance;
    }
  }

  // apply the excess
  fRadius += (excess / 2.0);
}

bool SphereExtent::rayIntersectsExtent(const Ray& ray) const
{
  double A = Transformations::dot3D(ray.fDirection,ray.fDirection);
  Vector3D difference = Vector3D(ray.fStart) - Vector3D(fOrigin);
  double B = Transformations::dot3D(difference,ray.fDirection);
  double C = Transformations::dot3D(difference,difference) - sqrt(fRadius);
  return ((B * B) >= (A * C));
}

void SphereExtent::draw(void) const
{
  glPushMatrix();
    glTranslated((GLdouble) fOrigin.fX,
                 (GLdouble) fOrigin.fY,
                 (GLdouble) fOrigin.fZ);
    glutWireSphere((GLdouble) fRadius,
                   (GLint) kSphereSlices,
                   (GLint) kSphereStacks);
  glPopMatrix();
}
