// ------------------------------
// Filename      : math.cpp
// Author        : Sven Maerivoet
// Last modified : 18/04/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#include "math.h"
#include <math.h>

// ----------
// class Math
// ----------

// --------------------------------------------
// initialization of static const class members
// --------------------------------------------
const double Math::kPi            = 3.14159265358979323846;
const double Math::kEpsilon       = 1E-12;
const double Math::kSolverEpsilon = 1E-50;
const double Math::kInfinity      = 1E100;

// -------
// methods
// -------
double Math::abs(double x)
{
  if (x >= 0.0) {
    return x;
  }
  else {
    return -x;
  }
}

double Math::sqr(double x)
{
  return (x * x);
}

double Math::cube(double x)
{
  return (x * x * x);
}

double Math::quart(double x)
{
  return (x * x * x * x);
}

double Math::cubicRoot(double x)
{
  if (x > 0.0) {
    return pow(x,1.0 / 3.0);
  }
  else if (x < 0.0) {
    return -pow(-x,1.0 / 3.0);
  }
  else {
    return 0.0;
  }
}

double Math::frac(double x)
{
  return (x - floor(x));
}

double Math::lerp(double a, double b, double fraction)
{
  return (a + (fraction * (b - a)));
}

double Math::arctan(double y, double x)
{
  double azimuth = kPi / 2.0;

  if (Math::abs(x) > kEpsilon) {
    azimuth = atan(y / x);
    if (x < 0.0) {
      azimuth += kPi;
    }
    else if (y < 0.0) {
      azimuth += (2.0 * kPi);
    }
  }
  else if (y < 0.0) {
    azimuth *= 3.0;
  }

  return azimuth;
}

double Math::radToDeg(double radians)
{
  return ((radians * 180.0) / kPi);
}

double Math::degToRad(double degrees)
{
  return ((degrees * kPi) / 180.0);
}

void Math::convertCartesianToSpherical(double x, double y, double z,
                                       double& radius, double& phi, double& theta)
{
  radius = sqrt(sqr(x) + sqr(y) + sqr(z));
  phi = arctan(y,x);
  theta = acos(z);
}

int Math::solveQuadraticEquation(double coefficients[3], double solutions[2])
{
  // the equation is described as :
  //   coefficients[0] +
  //   coefficients[1] * x +
  //   coefficients[2] * (x ^ 2)

  double a = coefficients[2];
  double b = -coefficients[1];
  double c = coefficients[0];

  // is it a 'linear' equation ?
  if (Math::abs(a) < kSolverEpsilon) {

    // is it a constant function ?
    if (Math::abs(b) < kSolverEpsilon) {

      if (Math::abs(c) < kSolverEpsilon) {
        // one real solution
        solutions[0] = 0.0;
        return 1;
      }
      else {
        // no real solution
        return 0;
      }
    }
    else {
      // one real solution
      solutions[0] = c / b;
      return 1;
    }
  }

  double discriminant = (b * b) - (4 * a * c);

  if (Math::abs(discriminant) < Math::kSolverEpsilon){
    // one real (double) solution
    solutions[0] = 0.5 * (b / a);
    return 1;
  }
  else if (discriminant < 0.0) {
    // no real solutions
    return 0;
  }

  // two real solutions
  discriminant = sqrt(discriminant);
  double t = 2.0 * a;

  solutions[0] = (b + discriminant) / t;
  solutions[1] = (b - discriminant) / t;

  return 2;
}

int Math::solveCubicEquation(double coefficients[4], double solutions[3])
{
  // the equation is described as :
  //   coefficients[0] +
  //   coefficients[1] * x +
  //   coefficients[2] * (x ^ 2) +
  //   coefficients[3] * (x ^ 3)

  // convert equation to normal form : x^3 + A*x^2 + B*x + C
  double A = coefficients[2] / coefficients[3];
  double B = coefficients[1] / coefficients[3];
  double C = coefficients[0] / coefficients[3];

  // substitute x = y - A/3 to eliminate the quadric term
  // so we obtain : x^3 + p*x + q
  double sqrA = A * A;
  double p = (1.0 / 3.0) * (((-1.0 / 3.0) * sqrA) + B);
  double q = (1.0 / 2.0) * (((2.0 / 27.0) * A * sqrA) - ((1.0 / 3.0) * A * B) + C);

  // use Cardano's formula
  double cubep = p * p * p;
  double discriminant = q * q + cubep;

  int nrOfSolutions = 0;

  if (Math::abs(discriminant) < kSolverEpsilon) {

    if (Math::abs(q) < kSolverEpsilon) {
      // one triple solution
      solutions[0] = 0.0;
      nrOfSolutions = 1;
    }
    else {
      // one single and one double solution
      double u = cubicRoot(-q);
      solutions[0] = 2 * u;
      solutions[1] = -u;
      nrOfSolutions = 2;
    }
  }
  else if (discriminant < 0.0) {
    // three real solutions
    double phi = (1.0 / 3.0) * acos(-q / sqrt(-cubep));
    double t = 2.0 * sqrt(-p);
    solutions[0] = t * cos(phi);
    solutions[1] = -t * cos(phi + (kPi / 3.0));
    solutions[2] = -t * cos(phi - (kPi / 3.0));
    nrOfSolutions = 3;
  }
  else {
    // one real solution
    double sqrtDiscriminant = sqrt(discriminant);
    double u = cubicRoot(sqrtDiscriminant - q);
    double v = -cubicRoot(sqrtDiscriminant + q);
    solutions[0] = u + v;
    nrOfSolutions = 1;
  }

  // resubstitute
  double substitute = (1.0 / 3.0) * A;

  for (int i = 0; i < nrOfSolutions; ++i) {
    solutions[i] -= substitute;
  }

  return nrOfSolutions;
}

int Math::solveQuarticEquation(double coefficients[5], double solutions[4])
{
  // the equation is described as :
  //   coefficients[0] +
  //   coefficients[1] * x +
  //   coefficients[2] * (x ^ 2) +
  //   coefficients[3] * (x ^ 3) +
  //   coefficients[4] * (x ^ 4)

  // convert equation to normal form : x^4 + A*x^3 + B*x^2 + C*x + D
  double A = coefficients[3] / coefficients[4];
  double B = coefficients[2] / coefficients[4];
  double C = coefficients[1] / coefficients[4];
  double D = coefficients[0] / coefficients[4];

  // substitute x = y - A/4 to eliminate the cubic term
  // so we obtain : x^4 + p*x^2 + q*x + r
  double sqrA = A * A;
  double p = -((3.0 / 8.0) * sqrA) + B;
  double q = ((1.0 / 8.0) * sqrA * A) - ((1.0 / 2.0) * A * B) + C;
  double r = -((3.0 / 256.0) * sqrA * sqrA) + ((1.0 / 16.0) * sqrA * B) - ((1.0 / 4.0) * A * C) + D;

  int nrOfSolutions = 0;
  double coeffs[4];

  if (Math::abs(r) < kSolverEpsilon) {

    // no absolute term y * (y^3 + p*y + q)
    coeffs[0] = q;
    coeffs[1] = p;
    coeffs[2] = 0.0;
    coeffs[3] = 1.0;
    nrOfSolutions = solveCubicEquation(coeffs,solutions);
    solutions[nrOfSolutions++] = 0.0;
  }
  else {
    // solve the resolvent cubic
    coeffs[0] = ((1.0 / 2.0) * r * p) - ((1.0 / 8.0) * q * q);
    coeffs[1] = -r;
    coeffs[2] = -(1.0 / 2.0) * p;
    coeffs[3] = 1.0;
    int dummy = solveCubicEquation(coeffs,solutions);

    // take the one real solution
    double z = solutions[0];

    // and build two quadric equations
    double u = (z * z) - r;
    double v = (2 * z) - p;

    if (Math::abs(u) < kSolverEpsilon) {
      u = 0.0;
    }
    else if (u > 0.0) {
      u = sqrt(u);
    }
    else {
      // no real solutions
      return 0;
    }

    if (Math::abs(v) < kSolverEpsilon) {
      v = 0.0;
    }
    else if (v > 0.0) {
      v = sqrt(v);
    }
    else {
      // no real solutions
      return 0;
    }

    coeffs[0] = z - u;
    if (q < 0.0) {
      coeffs[1] = -v;
    }
    else {
      coeffs[1] = v;
    }
    coeffs[2] = 1.0;

    nrOfSolutions = solveQuadraticEquation(coeffs,solutions);

    coeffs[0] = z + u;
    if (q < 0.0) {
      coeffs[1] = v;
    }
    else {
      coeffs[1] = -v;
    }
    coeffs[2] = 1.0;

    nrOfSolutions += solveQuadraticEquation(coeffs,solutions + nrOfSolutions);
  }

  // resubstitute
  double substitute = (1.0 / 4.0) * A;

  for (int i = 0; i < nrOfSolutions; ++i) {
    solutions[i] -= substitute;
  }

  return nrOfSolutions;
}
