// ------------------------------
// Filename      : opengl.h
// Author        : Sven Maerivoet
// Last modified : 03/06/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifndef SM_OPENGL_H // include-guard
#define SM_OPENGL_H

#include "constants.h"
#include "jitter.h"
#include "primitives.h"
#include "scene.h"
#include "ray.h"
#include "rgbapixmap.h"
#include "terrainmap.h"
#include <string>

using namespace std;

class TOpenGLApp {
  public:
    // some useful constants
    static const int  kQuasiFullScreen;
    static const bool kEnableDoubleBuffering;
    static const bool kDisableDoubleBuffering;
    static const bool kEnableDepthBuffer;
    static const bool kDisableDepthBuffer;
    static const bool kComputeShadows;
    static const bool kDontComputeShadows;
    static const bool kUseStochasticSampling;
    static const bool kDontUseStochasticSampling;
    static const bool kNormalizeColors;
    static const bool kDontNormalizeColors;
    static const bool kUseLandscapeMaterials;
    static const bool kDontUseLandscapeMaterials;
    static const bool kSmoothenVertexNormals;
    static const bool kDontSmoothenVertexNormals;
    static const bool kApplyTextureMapping;
    static const bool kDontApplyTextureMapping;

    // some useful colors
    static const Color kBlack;
    static const Color kBlue;
    static const Color kGreen;
    static const Color kCyan;
    static const Color kRed;
    static const Color kMagenta;
    static const Color kDarkBrown;
    static const Color kBrown;
    static const Color kLightBrown;
    static const Color kDarkGray;
    static const Color kGray;
    static const Color kLightGray;
    static const Color kWhite;
    static const Color kYellow;

    // the different possible lights
    static const int kLight1;
    static const int kLight2;
    static const int kLight3;
    static const int kLight4;
    static const int kLight5;
    static const int kLight6;
    static const int kLight7;
    static const int kLight8;

    // the different possible jitter-vectors
    enum EJitterVectors {kJitterVectors2,
                         kJitterVectors3,
                         kJitterVectors4,
                         kJitterVectors8,
                         kJitterVectors15,
                         kJitterVectors24,
                         kJitterVectors66};

    // the different anti-alias modes when ray tracing
    enum EAntiAliasMode {kUse1Sample,
                         kUse5Samples,
                         kUse9Samples};

    // the different shadow-computation modes when ray tracing
    enum EShadowComputationMode {kNoShadows,
                                 kHardShadows,
                                 kSoftShadows};

    // constructor
    TOpenGLApp(void);

    // destructor
    virtual ~TOpenGLApp(void);

    // initialization
    void initialize(int clientWindowXPos,
                    int clientWindowYPos,
                    int clientWindowXSize,
                    int clientWindowYSize,
                    string clientWindowTitle,
                    bool enableDoubleBuffering,
                    bool enableDepthBuffer);

    // explicit initialization (override if necessary)
    virtual void explicitInitialize(void);

    // callback registrations
    void registerDisplayFunc(void (*displayFuncPtr) (void)) const;
    void registerReshapeFunc(void (*reshapeFuncPtr) (int width, int height)) const;
    void registerKeyboardFunc(void (*keyboardFuncPtr) (unsigned char key, int x, int y)) const;
    void registerSpecialFunc(void (*specialFuncPtr) (int key, int x, int y)) const;
    void registerMouseFunc(void (*mouseFuncPtr) (int button, int state, int x, int y)) const;
    void registerMotionFunc(void (*motionFuncPtr) (int x, int y)) const;
    void registerPassiveMotionFunc(void (*passiveMotionFuncPtr) (int x, int y)) const;
    void registerIdleFunc(void (*idleFuncPtr) (void)) const;

    // callback-functions (override if necessary)

    // show a yellow cross on a blue background
    virtual void displayFunc(void);
    // maintain aspect-ratio of original world-window
    virtual void reshapeFunc(int width, int height);
    // enable projection management, lighting/shading control and miscellaneous
    virtual void keyboardFunc(unsigned char key, int x, int y);
    // enable standard camera control (move, slide, rotate, pitch and yaw)
    virtual void specialFunc(int key, int x, int y);
    virtual void mouseFunc(int button, int state, int x, int y);
    virtual void motionFunc(int x, int y);
    virtual void passiveMotionFunc(int x, int y);
    // enable continuous camera movements
    virtual void idleFunc(void);

    // activator for GLUT's event-handling
    void activateEventHandling(void) const;
 
    // the real display-callback (don't call directly !)
    void antiAliasedDisplayFunc(void);

    // miscellaneous
    void enableAntiAliasing(double jitterFactor, EJitterVectors nrOfJitterVectors);
    void disableAntiAliasing(void);
    void redisplay(void) const; // DON'T call this from inside a displayFunc()- or reshapeFunc()-override !
    int getClientWindowXSize(void) const;
    int getClientWindowYSize(void) const;
    void takeSnapshot(string filename) const;

    // OpenGL window-to-viewport mapping
    void setWorldWindow(double left, double top, double right, double bottom);
    void setViewport(int left, int top, int right, int bottom);
    Rectangle2D getWorldWindow(void) const;
    double getWorldWindowAspectRatio(void) const;
    Rectangle2D getViewport(void) const;

    // OpenGL projection management
    void setCamera(const Point3D& cameraEyePoint,
                   const Point3D& cameraViewPoint,
                   const Vector3D& cameraUpDirection);
    Point3D getCameraEyePoint(void) const;
    Point3D getCameraViewPoint(void) const;
    Vector3D getCameraUpDirection(void) const;
    void slideCamera(double du, double dv, double dn);
    void rollCamera(double degrees);
    void pitchCamera(double degrees);
    void yawCamera(double degrees);
    void setCameraPerspectiveShape(double cameraViewAngleInDegrees,
                                   double cameraAspectRatio,
                                   double cameraDistanceOfNearPlane,
                                   double cameraDistanceOfFarPlane);
    void setCameraPerspectiveShape(double cameraLeftPlane,
                                   double cameraTopPlane,
                                   double cameraRightPlane,
                                   double cameraBottomPlane,
                                   double cameraDistanceOfNearPlane,
                                   double cameraDistanceOfFarPlane);
    void setCameraObliqueShape(const Vector3D& projectionDirection,
                               double cameraLeftPlane,
                               double cameraTopPlane,
                               double cameraRightPlane,
                               double cameraBottomPlane,
                               double cameraDistanceOfNearPlane,
                               double cameraDistanceOfFarPlane);
    void setCameraOrthogonalShape(const Vector3D& projectionDirection,
                                  double cameraLeftPlane,
                                  double cameraTopPlane,
                                  double cameraRightPlane,
                                  double cameraBottomPlane,
                                  double cameraDistanceOfNearPlane,
                                  double cameraDistanceOfFarPlane);
    double getCameraViewAngle(void) const;
    double getCameraAspectRatio(void) const;
    double getCameraLeftPlane(void) const;
    double getCameraTopPlane(void) const;
    double getCameraRightPlane(void) const;
    double getCameraBottomPlane(void) const;
    double getCameraDistanceOfNearPlane(void) const;
    double getCameraDistanceOfFarPlane(void) const;
    void setCameraSlideStepSize(double cameraSlideStepSize);
    void setCameraRotateStepSize(double cameraRotateStepSizeInDegrees);
    void setCameraViewAngleStepSize(double cameraViewAngleStepSizeInDegrees);
    void setCameraAspectRatioStepSize(double cameraAspectRatioStepSize);
    void setCameraAdvanceStepSize(double cameraAdvanceStepSize);
    double getCameraSlideStepSize(void) const;
    double getCameraRotateStepSize(void) const;
    double getCameraViewAngleStepSize(void) const;
    double getCameraAspectRatioStepSize(void) const;
    double getCameraAdvanceStepSize(void) const;
    void loadCameraPath(string filename);
    bool advanceCameraAlongPath(void);
    void enableContinuousCameraMovements(void);
    void disableContinuousCameraMovements(void);
    void enableCameraMomentum(void);
    void disableCameraMomentum(void);
    void decoupleCameraYawAndRoll(void);
    void coupleCameraYawAndRoll(void);

    // OpenGL color, fog, lighting and texture management
    void clearScreen(void) const;
    void setBackgroundColor(double red, double green, double blue, double alpha) const;
    void setBackgroundColor(const Color& color) const;
    void setColor(const Color& color) const;
    void enableSmoothShading(void) const;
    void enableFlatShading(void) const;
    void setMaterial(const Color& ambientColor,
                     const Color& diffuseColor,
                     const Color& specularColor,
                     double shininessCoefficient) const;
    void setMaterialToBlackPlastic(void) const;
    void setMaterialToBrass(void) const;
    void setMaterialToBronze(void) const;
    void setMaterialToChrome(void) const;
    void setMaterialToCopper(void) const;
    void setMaterialToGold(void) const;
    void setMaterialToPewter(void) const;
    void setMaterialToSilver(void) const;
    void setMaterialToPolishedSilver(void) const;
    void setGlow(const Color& glowColor) const;
    void dontGlow(void) const;
    void setHardColor(const Color& color) const;
    void setGlobalAmbientLight(const Color& ambientColor) const;
    void enableExplicitViewpointCalculation(void) const;
    void disableExplicitViewpointCalculation(void) const;
    void enableLinearFog(const Color& fogColor, double start, double end) const;
    void enableExponentialFog(const Color& fogColor, double density) const;
    void disableFog(void) const;
    void switchLightingOn(void) const;
    void switchLightingOff(void) const;
    void switchLightOn(int light) const;
    void switchLightOff(int light) const;
    void setLightPosition(int light, const Point3D& position) const;
    void setLightDirection(int light, const Vector3D& direction) const;
    void setLightProperties(int light,
                            const Color& ambientColor,
                            const Color& diffuseColor,
                            const Color& specularColor) const;
    void makeSpotLight(int light,
                       const Vector3D& directionOfLightCone,
                       double cutOffAngle,
                       double exponent) const;
    void enableLightAttenuation(int light,
                                double constantAttenuation,
                                double linearAttenuation,
                                double quadraticAttenuation) const;
    void disableLightAttenuation(int light) const;
    int getUniqueTextureID(void) const;
    void enableGoodTextureInterpolation(void);
    void disableGoodTextureInterpolation(void);
    void enableGlowingTextures(void);
    void disableGlowingTextures(void);

    // 2D-graphics routines
    void point(double x, double y) const;
    void point(const Point2D& p) const;
    void setPointSize(double pointSize) const;
    void line(double x1, double y1, double x2, double y2) const;
    void line(const Point2D& p, const Point2D& q) const;
    void lineTo(double x, double y);
    void lineTo(const Point2D& p);
    void moveTo(double x, double y);
    void moveTo(const Point2D& p);
    void lineRel(double dx, double dy);
    void moveRel(double dx, double dy);
    void forward(double distance, bool isVisible);
    void turnTo(double degrees);
    void turn(double degrees);
    void ngon(int n, double x, double y, double radius, double startAngleInDegrees);
    void ngon(int n, const Point2D& center, double radius, double startAngleInDegrees);
    void circle(double x, double y, double radius);
    void circle(const Point2D& center, double radius);
    void polyLineFromFile(string filename) const;
    void drawString(int x, int y, string str, const Color& color);

    // 3D-graphics routines
    void point(double x, double y, double z) const;
    void point(const Point3D& p) const;
    void line(double x1, double y1, double z1,
              double x2, double y2, double z2) const;
    void line(const Point3D& p, const Point3D& q) const;
    void lineTo(double x, double y, double z);
    void lineTo(const Point3D& p);
    void moveTo(double x, double y, double z);
    void moveTo(const Point3D& p);
    void lineRel(double dx, double dy, double dz);
    void moveRel(double dx, double dy, double dz);
    void cone() const;
    void cube() const;
    void cylinder() const;
    void dodecahedron() const;
    void icosahedron() const;
    void octahedron() const;
    void sphere() const;
    void taperedCylinder(double topRadius) const;
    void teapot() const;
    void tetrahedron() const;
    void torus(double tubeRadius, double torusRadius) const;

    // miscellaneous 3D-graphics routines
    void textureQuad(const RGBApixmap* const pixmap,
                     const Point2D& tUpperLeft,
                     const Point2D& tLowerLeft,
                     const Point2D& tLowerRight,
                     const Point2D& tUpperRight,
                     const Point3D& pUpperLeft,
                     const Point3D& pLowerLeft,
                     const Point3D& pLowerRight,
                     const Point3D& pUpperRight) const;
    void drawLandscape(const TerrainMap& terrainMap,
                       bool useLandscapeMaterials,
                       bool smoothenVertexNormals,
                       bool applyTextureMapping) const;
    void showAxes(double length, double thickness) const;

    // scene-rendering (OpenGL and ray tracing)
    void loadScene(string filename);
    void clearScene(void);
    void setSceneViewSettings(void);
    void enableSceneLighting(void) const;
    void drawScene(void);
    void rayTraceScene(int blockSize,
                       int nrOfConeSamples,
                       EShadowComputationMode shadowComputationMode,
                       double shadowFeelerJitterStrength,
                       EAntiAliasMode antiAliasMode,
                       bool useStochasticSampling,
                       double jitterStrength,
                       bool normalizeColors,
                       double normalizeBrightness,
                       bool findNormals,
                       double findNormalsProbability,
                       bool followRays,
                       double followRaysProbability);
    long getNrOfRaysShot(void);

    // 2D- and 3D-transformations
    void initCT(void) const;
    void rotate2D(double degrees) const;
    void rotate3D(double degrees, double x, double y, double z) const;
    void translate2D(double dx, double dy) const;
    void translate2D(const Point2D& p) const;
    void translate3D(double dx, double dy, double dz) const;
    void translate3D(const Point3D& p) const;
    void scale2D(double sx, double sy) const;
    void scale3D(double sx, double sy, double sz) const;
    void pushCT(void) const;
    void popCT(void) const;

  protected:
    // utility methods
    static string convertDoubleToString(double x);

    // data-members
    bool fScheduleRedisplay;

  private:
    // copy-constructor (can't be used !)
    TOpenGLApp(const TOpenGLApp& openGLApp);

    // utility methods
    void finishFrame(void) const; // DON'T call this from inside a displayFunc()- or reshapeFunc()-override !
    void setModelViewMatrix(void) const;
    void rotateCameraAroundAxes(Vector3D& a, Vector3D& b, double degrees) const;
    Color samplePixel(EAntiAliasMode antiAliasMode,
                      bool useStochasticSampling,
                      int row,
                      int column,
                      int nrOfConeSamples,
                      int blockSize,
                      double pixelWidth,
                      double pixelHeight,
                      double jitterStrength);

    // the different continuous camera-moves
    enum ECameraMoves {kCameraStill,
                       kCameraMoving,
                       kCameraSliding,
                       kCameraRotating,
                       kCameraPitching,
                       kCameraYawing,
                       kMoveXLeft,kMoveXRight,
                       kMoveYDown,kMoveYUp,
                       kMoveZForward,kMoveZBackward,
                       kSlideXLeft,kSlideXRight,
                       kSlideYDown,kSlideYUp,
                       kSlideZForward,kSlideZBackward,
                       kRotateLeft,kRotateRight,
                       kPitchDown,kPitchUp,
                       kYawLeft,kYawRight};

    // data-members
    int            fClientWindowXPos;
    int            fClientWindowYPos;
    int            fClientWindowXSize;
    int            fClientWindowYSize;
    int            fOldClientWindowXPos;
    int            fOldClientWindowYPos;
    int            fOldClientWindowXSize;
    int            fOldClientWindowYSize;
    string         fClientWindowTitle;
    bool           fFullScreenMode;
    bool           fEnableDoubleBuffering;
    bool           fEnableDepthBuffer;
    bool           fAntiAliasingEnabled;
    double         fJitterFactor;
    int            fNrOfJitterVectors;
    JitterVectors* fJitterVectors;
    Rectangle2D    fWorldWindow;
    double         fWorldWindowAspectRatio;
    Rectangle2D    fViewport;
    Point3D        fCurrentPosition;
    double         fCurrentDirection;
    Scene          fScene;
    Point3D        fCameraEyePoint;
    Point3D        fCameraViewPoint;
    Vector3D       fCameraUpDirection;
    Vector3D       fCameraU;
    Vector3D       fCameraV;
    Vector3D       fCameraN;
    double         fCameraViewAngle;
    double         fCameraAspectRatio;
    double         fCameraDistanceOfNearPlane;
    double         fCameraDistanceOfFarPlane;
    double         fCameraLeftPlane;
    double         fCameraTopPlane;
    double         fCameraRightPlane;
    double         fCameraBottomPlane;
    double         fCameraSlideStepSize;
    double         fCameraRotateStepSize;
    double         fCameraViewAngleStepSize;
    double         fCameraAspectRatioStepSize;
    double         fCameraAdvanceStepSize;
    int            fCameraNrOfControlPointsOnPath;
    Point3D*       fCameraControlPointsOnPath;
    double         fCameraDistanceOnSegmentTraveled;
    int            fCameraLastControlPointVisited;
    bool           fContinuousCameraMovements;
    bool           fCameraHasMomentum;
    bool           fCameraYawAndRollCoupled;
    ECameraMoves   fCurrentCameraMovement;
    ECameraMoves   fCameraMove;
    ECameraMoves   fCameraSlide;
    ECameraMoves   fCameraRotate;
    ECameraMoves   fCameraPitch;
    ECameraMoves   fCameraYaw;
    bool           fGoodTextureInterpolation;
    bool           fGlowingTextures;
    int            fSnapshotSequenceNr;
    Ray*           fRays;
};

#endif
