// ------------------------------
// Filename      : primitives.cpp
// Author        : Sven Maerivoet
// Last modified : 20/04/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#include "constants.h"
#include "primitives.h"
#include "math.h"
#include <math.h>
#include <stdlib.h>

using namespace std;

// -------------
// class Point2D
// -------------

// ------------
// constructors
// ------------
Point2D::Point2D(void) : fX(0.0),
                         fY(0.0)
{
}

Point2D::Point2D(double x, double y) : fX(x),
                                       fY(y)
{
}

Point2D::Point2D(const Point2D& point2D) : fX(point2D.fX),
                                           fY(point2D.fY)
{
}

// ----------
// destructor
// ----------
Point2D::~Point2D(void)
{
}

// -------
// methods
// -------
void Point2D::set(double x, double y)
{
  fX = x;
  fY = y;
}

// -------------
// class Point3D
// -------------

// ------------
// constructors
// ------------
Point3D::Point3D(void) : Point2D(),
                         fZ(0.0)
{
}

Point3D::Point3D(double x, double y, double z) : Point2D(x,y),
                                                 fZ(z)
{
}

Point3D::Point3D(const Point3D& point3D) : Point2D(point3D),
                                           fZ(point3D.fZ)
{
}

// ----------
// destructor
// ----------
Point3D::~Point3D(void)
{
}

// -------
// methods
// -------
void Point3D::set(double x, double y, double z)
{
  fX = x;
  fY = y;
  fZ = z;
}

void Point3D::build4tuple(double v[]) const
{
  v[0] = fX;
  v[1] = fY;
  v[2] = fZ;
  v[3] = 1.0;
}

void Point3D::build4tuple(float v[]) const
{
  v[0] = (float) fX;
  v[1] = (float) fY;
  v[2] = (float) fZ;
  v[3] = 1.0f;
}

// -----------------
// class Rectangle2D
// -----------------

// ------------
// constructors
// ------------
Rectangle2D::Rectangle2D(void) : fLeft(0.0),
                                 fTop(0.0),
                                 fRight(0.0),
                                 fBottom(0.0)
{
}

Rectangle2D::Rectangle2D(double left, double top, double right, double bottom) : fLeft(left),
                                                                                 fTop(top),
                                                                                 fRight(right),
                                                                                 fBottom(bottom)
{
}

Rectangle2D::Rectangle2D(const Rectangle2D& rectangle2D) : fLeft(rectangle2D.fLeft),
                                                           fTop(rectangle2D.fTop),
                                                           fRight(rectangle2D.fRight),
                                                           fBottom(rectangle2D.fBottom)
{
}

// ----------
// destructor
// ----------
Rectangle2D::~Rectangle2D(void)
{
}

// -------
// methods
// -------
void Rectangle2D::set(double left, double top, double right, double bottom)
{
  fLeft = left;
  fTop = top;
  fRight = right;
  fBottom = bottom;
}

double Rectangle2D::getWidth(void) const
{
  return Math::abs(fLeft - fRight);
}

double Rectangle2D::getHeight(void) const
{
  return Math::abs(fTop - fBottom);
}

double Rectangle2D::getAspectRatio(void) const
{
  return (getWidth() / getHeight());
}

// --------------
// class Vector2D
// --------------

// ------------
// constructors
// ------------
Vector2D::Vector2D(void) : fX(0.0),
                           fY(0.0)
{
}

Vector2D::Vector2D(double x, double y) : fX(x),
                                         fY(y)
{
}

Vector2D::Vector2D(const Point2D& point2D) : fX(point2D.fX),
                                             fY(point2D.fY)
{
}

Vector2D::Vector2D(const Vector2D& vector2D) : fX(vector2D.fX),
                                               fY(vector2D.fY)
{
}

// ----------
// destructor
// ----------
Vector2D::~Vector2D(void)
{
}

// -----------------
// vector arithmetic
// -----------------
Vector2D& Vector2D::operator=(const Vector2D& vector2D)
{
  // don't assign a vector to itself
  if (this != &vector2D) {
  	set(vector2D.fX,vector2D.fY);
  }

  return *this;
}

double Vector2D::length(void) const
{
  return sqrt((fX * fX) + (fY * fY));
}

void Vector2D::normalize(void)
{
  double vectorLength = length();

  if (vectorLength != 0.0) {
  	set(fX / vectorLength,
  	    fY / vectorLength);
  }
}

void Vector2D::flip(void)
{
  fX = -fX;
  fY = -fY;
}

Vector2D Vector2D::operator+(const Vector2D& vector2D) const
{
  Vector2D v(fX + vector2D.fX,
             fY + vector2D.fY);
  return v;
}

Vector2D& Vector2D::operator+=(const Vector2D& vector2D)
{
  fX += vector2D.fX;
  fY += vector2D.fY;
  return *this;
}

Vector2D Vector2D::operator-(const Vector2D& vector2D) const
{
  Vector2D v(fX - vector2D.fX,
             fY - vector2D.fY);
  return v;
}

Vector2D& Vector2D::operator-=(const Vector2D& vector2D)
{
  fX -= vector2D.fX;
  fY -= vector2D.fY;
  return *this;
}

Vector2D Vector2D::operator*(double scalar) const
{
  Vector2D v(fX * scalar,
             fY * scalar);
  return v;
}

Vector2D& Vector2D::operator*=(double scalar)
{
  fX /= scalar;
  fY /= scalar;
  return *this;
}

Vector2D Vector2D::operator/(double scalar) const
{
  Vector2D v(fX / scalar,
             fY / scalar);
  return v;
}

Vector2D& Vector2D::operator/=(double scalar)
{
  fX /= scalar;
  fY /= scalar;
  return *this;
}

double Vector2D::inner(const Vector2D& vector2D) const
{
  return ((fX * vector2D.fX) +
          (fY * vector2D.fY));
}

double Vector2D::angle(const Vector2D& vector2D) const
{
  // make a local copy of this vector
  Vector2D v = *this;

  // make a local copy of the other vector
  Vector2D w = vector2D;

  // normalize the two vectors
  v.normalize();
  w.normalize();

  return acos(v.inner(w));
}

// -------
// methods
// -------
void Vector2D::set(double x, double y)
{
  fX = x;
  fY = y;
}

// --------------
// class Vector3D
// --------------

// ------------
// constructors
// ------------
Vector3D::Vector3D(void) : fX(0.0),
                           fY(0.0),
                           fZ(0.0)
{
}

Vector3D::Vector3D(double x, double y, double z) : fX(x),
                                                   fY(y),
                                                   fZ(z)
{
}

Vector3D::Vector3D(const Point3D& point3D) : fX(point3D.fX),
                                             fY(point3D.fY),
                                             fZ(point3D.fZ)
{
}

Vector3D::Vector3D(const Vector3D& vector3D) : fX(vector3D.fX),
                                               fY(vector3D.fY),
                                               fZ(vector3D.fZ)
{
}

// ----------
// destructor
// ----------
Vector3D::~Vector3D(void)
{
}

// -----------------
// vector arithmetic
// -----------------
Vector3D& Vector3D::operator=(const Vector3D& vector3D)
{
  // don't assign a vector to itself
  if (this != &vector3D) {
  	set(vector3D.fX,vector3D.fY,vector3D.fZ);
  }

  return *this;
}

double Vector3D::length(void) const
{
  return sqrt((fX * fX) + (fY * fY) + (fZ * fZ));
}

void Vector3D::normalize(void)
{
  double vectorLength = length();

  if (vectorLength != 0.0) {
  	set(fX / vectorLength,
  	    fY / vectorLength,
  	    fZ / vectorLength);
  }
}

void Vector3D::flip(void)
{
  fX = -fX;
  fY = -fY;
  fZ = -fZ;
}

Vector3D Vector3D::operator+(const Vector3D& vector3D) const
{
  Vector3D v(fX + vector3D.fX,
             fY + vector3D.fY,
             fZ + vector3D.fZ);
  return v;
}

Vector3D& Vector3D::operator+=(const Vector3D& vector3D)
{
  fX += vector3D.fX;
  fY += vector3D.fY;
  fZ += vector3D.fZ;
  return *this;
}

Vector3D Vector3D::operator-(const Vector3D& vector3D) const
{
  Vector3D v(fX - vector3D.fX,
             fY - vector3D.fY,
             fZ - vector3D.fZ);
  return v;
}

Vector3D& Vector3D::operator-=(const Vector3D& vector3D)
{
  fX -= vector3D.fX;
  fY -= vector3D.fY;
  fZ -= vector3D.fZ;
  return *this;
}

Vector3D Vector3D::operator*(double scalar) const
{
  Vector3D v(fX * scalar,
             fY * scalar,
             fZ * scalar);
  return v;
}

Vector3D& Vector3D::operator*=(double scalar)
{
  fX *= scalar;
  fY *= scalar;
  fZ *= scalar;
  return *this;
}

Vector3D Vector3D::operator/(double scalar) const
{
  Vector3D v(fX / scalar,
             fY / scalar,
             fZ / scalar);
  return v;
}

Vector3D& Vector3D::operator/=(double scalar)
{
  fX /= scalar;
  fY /= scalar;
  fZ /= scalar;
  return *this;
}

double Vector3D::inner(const Vector3D& vector3D) const
{
  return ((fX * vector3D.fX) +
          (fY * vector3D.fY) +
          (fZ * vector3D.fZ));
}

Vector3D Vector3D::outer(const Vector3D& vector3D) const
{
  double x = (fY * vector3D.fZ) - (fZ * vector3D.fY);
  double y = (fZ * vector3D.fX) - (fX * vector3D.fZ);
  double z = (fX * vector3D.fY) - (fY * vector3D.fX);

  return Vector3D(x,y,z);
}

double Vector3D::angle(const Vector3D& vector3D) const
{
  // make a local copy of this vector
  Vector3D v = *this;

  // make a local copy of the other vector
  Vector3D w = vector3D;

  // normalize the two vectors
  v.normalize();
  w.normalize();

  return acos(v.inner(w));
}

// -------
// methods
// -------
void Vector3D::set(double x, double y, double z)
{
  fX = x;
  fY = y;
  fZ = z;
}

Vector3D Vector3D::calcNewellNormal3(const Point3D& p1,
                                     const Point3D& p2,
                                     const Point3D& p3)
{
  Vector3D normal;

  // the normal is approximated using Martin Newell's method
  normal.fX = ((p1.fY - p2.fY) * (p1.fZ + p2.fZ)) +
              ((p2.fY - p3.fY) * (p2.fZ + p3.fZ)) +
              ((p3.fY - p1.fY) * (p3.fZ + p1.fZ));

  normal.fY = ((p1.fZ - p2.fZ) * (p1.fX + p2.fX)) +
              ((p2.fZ - p3.fZ) * (p2.fX + p3.fX)) +
              ((p3.fZ - p1.fZ) * (p3.fX + p1.fX));

  normal.fZ = ((p1.fX - p2.fX) * (p1.fY + p2.fY)) +
              ((p2.fX - p3.fX) * (p2.fY + p3.fY)) +
              ((p3.fX - p1.fX) * (p3.fY + p1.fY));

  return normal;
}

Vector3D Vector3D::calcNewellNormal4(const Point3D& p1,
                                     const Point3D& p2,
                                     const Point3D& p3,
                                     const Point3D& p4)
{
  Vector3D normal;

  // the normal is approximated using Martin Newell's method
  normal.fX = ((p1.fY - p2.fY) * (p1.fZ + p2.fZ)) +
              ((p2.fY - p3.fY) * (p2.fZ + p3.fZ)) +
              ((p3.fY - p4.fY) * (p3.fZ + p4.fZ)) +
              ((p4.fY - p1.fY) * (p4.fZ + p1.fZ));

  normal.fY = ((p1.fZ - p2.fZ) * (p1.fX + p2.fX)) +
              ((p2.fZ - p3.fZ) * (p2.fX + p3.fX)) +
              ((p3.fZ - p4.fZ) * (p3.fX + p4.fX)) +
              ((p4.fZ - p1.fZ) * (p4.fX + p1.fX));

  normal.fZ = ((p1.fX - p2.fX) * (p1.fY + p2.fY)) +
              ((p2.fX - p3.fX) * (p2.fY + p3.fY)) +
              ((p3.fX - p4.fX) * (p3.fY + p4.fY)) +
              ((p4.fX - p1.fX) * (p4.fY + p1.fY));

  return normal;
}

void Vector3D::build4tuple(double v[]) const
{
  v[0] = fX;
  v[1] = fY;
  v[2] = fZ;
  v[3] = 1.0;
}

void Vector3D::build4tuple(float v[]) const
{
  v[0] = (float) fX;
  v[1] = (float) fY;
  v[2] = (float) fZ;
  v[3] = 1.0f;
}

// -----------
// class Color
// -----------

// ------------
// constructors
// ------------
Color::Color(void) : fRed(0.0),
                     fGreen(0.0),
                     fBlue(0.0)
{
}

Color::Color(double red, double green, double blue) : fRed(red),
                                                      fGreen(green),
                                                      fBlue(blue)
{
}

Color::Color(const Color& color) : fRed(color.fRed),
                                   fGreen(color.fGreen),
                                   fBlue(color.fBlue)
{
}

// ----------
// destructor
// ----------
Color::~Color(void)
{
}

// ----------------
// color arithmetic
// ----------------
Color& Color::operator=(const Color& color)
{
  // don't assign a color to itself
  if (this != &color) {
  	set(color.fRed,color.fGreen,color.fBlue);
  }

  return *this;
}

Color Color::operator+(const Color& color) const
{
  return Color(fRed + color.fRed,
               fGreen + color.fGreen,
               fBlue + color.fBlue);
}

Color Color::operator+(double value) const
{
  return Color(fRed + value,fGreen + value,fBlue + value);
}

Color& Color::operator+=(const Color& color)
{
  fRed += color.fRed;
  fGreen += color.fGreen;
  fBlue += color.fBlue;
  return *this;
}

Color& Color::operator+=(double value)
{
  fRed += value;
  fGreen += value;
  fBlue += value;
  return *this;
}

Color Color::operator-(const Color& color) const
{
  return Color(fRed - color.fRed,
               fGreen - color.fGreen,
               fBlue - color.fBlue);
}

Color Color::operator-(double value) const
{
  return Color(fRed - value,fGreen - value,fBlue - value);
}

Color& Color::operator-=(const Color& color)
{
  fRed -= color.fRed;
  fGreen -= color.fGreen;
  fBlue -= color.fBlue;
  return *this;
}

Color& Color::operator-=(double value)
{
  fRed -= value;
  fGreen -= value;
  fBlue -= value;
  return *this;
}

Color Color::operator*(const Color& color) const
{
  return Color(fRed * color.fRed,
               fGreen * color.fGreen,
               fBlue * color.fBlue);
}

Color Color::operator*(double value) const
{
  return Color(fRed * value,fGreen * value,fBlue * value);
}

Color& Color::operator*=(const Color& color)
{
  fRed *= color.fRed;
  fGreen *= color.fGreen;
  fBlue *= color.fBlue;
  return *this;
}

Color& Color::operator*=(double value)
{
  fRed *= value;
  fGreen *= value;
  fBlue *= value;
  return *this;
}

Color Color::operator/(const Color& color) const
{
  return Color(fRed / color.fRed,
               fGreen / color.fGreen,
               fBlue / color.fBlue);
}

Color Color::operator/(double value) const
{
  return Color(fRed / value,fGreen * value,fBlue * value);
}

Color& Color::operator/=(const Color& color)
{
  fRed /= color.fRed;
  fGreen /= color.fGreen;
  fBlue /= color.fBlue;
  return *this;
}

Color& Color::operator/=(double value)
{
  fRed /= value;
  fGreen /= value;
  fBlue /= value;
  return *this;
}

double Color::calcGrayLevel(const Color& color) const
{
  return ((0.30 * color.fRed) +
          (0.59 * color.fGreen) +
          (0.11 * color.fBlue));
}

Color Color::calcGrayLevelColor(const Color& color) const
{
  double grayLevel = calcGrayLevel(color);
  return Color(grayLevel,grayLevel,grayLevel);
}

void Color::clamp(void)
{
  // clamp all RGB-values in the [0,1] range
  if (fRed < 0.0) {
    fRed = 0.0;
  }
  else if (fRed > 1.0) {
    fRed = 1.0;
  }

  if (fGreen < 0.0) {
    fGreen = 0.0;
  }
  else if (fGreen > 1.0) {
    fGreen = 1.0;
  }

  if (fBlue < 0.0) {
    fBlue = 0.0;
  }
  else if (fBlue > 1.0) {
    fBlue = 1.0;
  }
}

// -------
// methods
// -------
void Color::set(double red, double green, double blue)
{
  fRed = red;
  fGreen = green;
  fBlue = blue;
}

void Color::build4tuple(double v[]) const
{
  v[0] = fRed;
  v[1] = fGreen;
  v[2] = fBlue;
  v[3] = 1.0;
}

void Color::build4tuple(float v[]) const
{
  v[0] = (float) fRed;
  v[1] = (float) fGreen;
  v[2] = (float) fBlue;
  v[3] = 1.0f;
}
