// ------------------------------
// Filename      : ray.cpp
// Author        : Sven Maerivoet
// Last modified : 05/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#include "math.h"
#include "ray.h"
#include "rng.h"
#include <math.h>

using namespace std;

// ---------
// class Ray
// ---------

// ------------
// constructors
// ------------
Ray::Ray(void)
{
  clear();
}

Ray::Ray(const Point3D& start)
{
  clear();
  fStart = start;
}

Ray::Ray(const Point3D& start, const Vector3D& direction)
{
  clear();
  fStart = start;
  fDirection = direction;
}

Ray::Ray(const Ray& ray)
{
  copyFromRay(ray);
}

// ----------
// destructor
// ----------
Ray::~Ray(void)
{
}

// -------
// methods
// -------
void Ray::clear(void)
{
  fStart = Point3D(0.0,0.0,0.0);
  fDirection = Vector3D(0.0,0.0,0.0);
  fRecurseLevel = 1;
  fRow = 0;
  fColumn = 0;
  fIsReflective = false;
  fIsRefractive = false;
  fNrOfObjectsPierced = 0;
  fNrOfConeSamples = 1;
  fSampleWeight = 1.0;
  fFollowRay = false;
  fColor = Color(0.0,0.0,0.0);
}

void Ray::copyObjectsPierced(const Ray& ray)
{
  fNrOfObjectsPierced = ray.fNrOfObjectsPierced;
  for (int i = 0; i < kMaxObjectsPierced; ++i) {
    fObjectsPierced[i] = ray.fObjectsPierced[i];
  }
}

void Ray::copyFromRay(const Ray& ray)
{
  fStart = ray.fStart;
  fDirection = ray.fDirection;
  fRecurseLevel = ray.fRecurseLevel;
  fRow = ray.fRow;
  fColumn = ray.fColumn;
  fIsReflective = ray.fIsReflective;
  fIsRefractive = ray.fIsRefractive;
  copyObjectsPierced(ray);
  fNrOfConeSamples = ray.fNrOfConeSamples;
  fSampleWeight = ray.fSampleWeight;
  fFollowRay = ray.fFollowRay;
  fColor = ray.fColor;
}

void Ray::jitterRaysInCone(vector<Ray>& jitteredRays,
                           const Ray& centralConeRay,
                           double coneRadius)
{
  jitteredRays.clear();

  Ray centerRay(centralConeRay);
  centerRay.fDirection.normalize();
  centerRay.fSampleWeight = 1.0 / centerRay.fNrOfConeSamples;
  jitteredRays.push_back(centerRay);

  // calculate the center-ray's direction in spherical coordinates
  double radius, phi, theta;
  Math::convertCartesianToSpherical(centerRay.fDirection.fX,
                                    centerRay.fDirection.fY,
                                    centerRay.fDirection.fZ,
                                    radius,phi,theta);

  // precalculate some goniometric terms
  double cosPhi = cos(phi);
  double sinPhi = sin(phi);
  double cosTheta = cos(theta);
  double sinTheta = sin(theta);

  double sampleWeight = 1.0 / centerRay.fNrOfConeSamples;
  double jitterStrength = coneRadius / kJitterStrengthFactor;

  for (int coneSample = 0; coneSample < (centerRay.fNrOfConeSamples - 1); ++coneSample) {
    Ray jitteredRay(centerRay);

    // determine the ray's position in the discretized cone
    Vector3D jitteredDirection(coneRadius * (cos(coneSample * ((2.0 * Math::kPi) / (centerRay.fNrOfConeSamples - 1)))),
                               coneRadius * (sin(coneSample * ((2.0 * Math::kPi) / (centerRay.fNrOfConeSamples - 1)))),
                               1.0);

    // jitter this position
    jitteredDirection.fX += -(jitterStrength / 2.0) + (jitterStrength * RNG::uniform());
    jitteredDirection.fY += -(jitterStrength / 2.0) + (jitterStrength * RNG::uniform());

    // rotate this jittered ray so it lies 'in the direction' of the cone's central ray
    jitteredRay.fDirection = Vector3D((( cosPhi   * cosTheta * jitteredDirection.fX) +
                                       (-sinPhi   *            jitteredDirection.fY) +
                                       ( cosPhi   * sinTheta * jitteredDirection.fZ)),
                                      (( sinPhi   * cosTheta * jitteredDirection.fX) +
                                       ( cosPhi   *            jitteredDirection.fY) +
                                       ( sinPhi   * sinTheta * jitteredDirection.fZ)),
                                      ((-sinTheta *            jitteredDirection.fX) +
                                       ( cosTheta *            jitteredDirection.fZ)));
    jitteredRay.fSampleWeight = sampleWeight;

    jitteredRays.push_back(jitteredRay);
  }
}

// -------------
// class HitInfo
// -------------

// -----------
// constructor
// -----------
HitInfo::HitInfo(void) : fHitTime(-Math::kInfinity),
                         fObjectHit(0),
                         fSurface(0),
                         fIsEntering(false),
                         fHitPoint(0.0,0.0,0.0),
			 fBarycentricU(0.0),
			 fBarycentricV(0.0),
                         fHitNormal(0.0,0.0,0.0)
{
}

// ----------
// destructor
// ----------
HitInfo::~HitInfo(void)
{
}

// -------
// methods
// -------
void HitInfo::set(const HitInfo& hitInfo)
{
  fHitTime = hitInfo.fHitTime;
  fObjectHit = hitInfo.fObjectHit;
  fSurface = hitInfo.fSurface;
  fIsEntering = hitInfo.fIsEntering;
  fHitPoint = hitInfo.fHitPoint;
  fBarycentricU = hitInfo.fBarycentricU;
  fBarycentricV = hitInfo.fBarycentricV;
  fHitNormal = hitInfo.fHitNormal;
}

// ------------------
// class Intersection
// ------------------

// -----------
// constructor
// -----------
Intersection::Intersection(void) : fNrOfHits(0)
{
}

// ----------
// destructor
// ----------
Intersection::~Intersection(void)
{
}

// -------
// methods
// -------
void Intersection::set(const Intersection& intersection)
{
  fNrOfHits = intersection.fNrOfHits;
  for (int i = 0; i < kMaxHits; ++i) {
    fHits[i].set(intersection.fHits[i]);
  }
}

bool Intersection::grazingAtSurface(void)
{
  // note that this is maybe incorrect for shapes like pure planes and polygons
  return (fNrOfHits == 1);
}

// ---------------------
// class Transformations
// ---------------------

// -------
// methods
// -------
void Transformations::transformPoint(Point3D& result,
                                     const AffineTransformation& affine,
                                     const Point3D& point)
{
  result.fX = affine.m[0] * point.fX +
              affine.m[4] * point.fY +
              affine.m[8] * point.fZ +
              affine.m[12];
  result.fY = affine.m[1] * point.fX +
              affine.m[5] * point.fY +
              affine.m[9] * point.fZ +
              affine.m[13];
  result.fZ = affine.m[2]  * point.fX +
              affine.m[6]  * point.fY +
              affine.m[10] * point.fZ +
              affine.m[14];
}

void Transformations::translatePoint(Point3D& result,
                                     const AffineTransformation& affine,
                                     const Point3D& point)
{
  result.fX = point.fX + affine.m[12];
  result.fY = point.fY + affine.m[13];
  result.fZ = point.fZ + affine.m[14];
}

void Transformations::transformVector(Vector3D& result,
                                      const AffineTransformation& affine,
                                      const Vector3D& vec)
{
  result.fX = affine.m[0] * vec.fX +
              affine.m[4] * vec.fY +
              affine.m[8] * vec.fZ;
  result.fY = affine.m[1] * vec.fX +
              affine.m[5] * vec.fY +
              affine.m[9] * vec.fZ;
  result.fZ = affine.m[2]  * vec.fX +
              affine.m[6]  * vec.fY +
              affine.m[10] * vec.fZ;
}


void Transformations::transformNormal(Vector3D& result,
                                      const AffineTransformation& affine,
                                      const Vector3D& normal)
{
  result.fX = affine.m[0] * normal.fX +
              affine.m[1] * normal.fY +
              affine.m[2] * normal.fZ;
  result.fY = affine.m[4] * normal.fX +
              affine.m[5] * normal.fY +
              affine.m[6] * normal.fZ;
  result.fZ = affine.m[8]  * normal.fX +
              affine.m[9]  * normal.fY +
              affine.m[10] * normal.fZ;
}

void Transformations::transformRay(Ray& result,
                                   const AffineTransformation& affine,
                                   const Ray& ray)
{
  result.fStart.fX = affine.m[0] * ray.fStart.fX +
                     affine.m[4] * ray.fStart.fY +
                     affine.m[8] * ray.fStart.fZ +
                     affine.m[12];
  result.fStart.fY = affine.m[1] * ray.fStart.fX +
                     affine.m[5] * ray.fStart.fY +
                     affine.m[9] * ray.fStart.fZ +
                     affine.m[13];
  result.fStart.fZ = affine.m[2]  * ray.fStart.fX +
                     affine.m[6]  * ray.fStart.fY +
                     affine.m[10] * ray.fStart.fZ +
                     affine.m[14];

  result.fDirection.fX = affine.m[0] * ray.fDirection.fX +
                         affine.m[4] * ray.fDirection.fY +
                         affine.m[8] * ray.fDirection.fZ;
  result.fDirection.fY = affine.m[1] * ray.fDirection.fX +
                         affine.m[5] * ray.fDirection.fY +
                         affine.m[9] * ray.fDirection.fZ;
  result.fDirection.fZ = affine.m[2]  * ray.fDirection.fX +
                         affine.m[6]  * ray.fDirection.fY +
                         affine.m[10] * ray.fDirection.fZ;
}

Point3D Transformations::calcRayPosition(const Point3D& start,
                                         const Vector3D& direction,
                                         double distance)
{
  return Point3D(start.fX + (direction.fX * distance),
                 start.fY + (direction.fY * distance),
                 start.fZ + (direction.fZ * distance));
}

Point3D Transformations::calcRayPosition(const Ray& ray,
                                         double distance)
{
  return calcRayPosition(ray.fStart,ray.fDirection,distance);
}

double Transformations::dot3D(const Vector3D& a, const Vector3D& b)
{
  return ((a.fX * b.fX) +
          (a.fY * b.fY) +
          (a.fZ * b.fZ));
}

double Transformations::dot3D(const Point3D& a, const Vector3D& b)
{
  return ((a.fX * b.fX) +
          (a.fY * b.fY) +
          (a.fZ * b.fZ));
}

double Transformations::dot3D(const Point3D& a, const Point3D& b)
{
  return ((a.fX * b.fX) +
          (a.fY * b.fY) +
          (a.fZ * b.fZ));
}
