// ------------------------------
// Filename      : ray.h
// Author        : Sven Maerivoet
// Last modified : 20/04/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifndef SM_RAY_H // include-guard
#define SM_RAY_H

#ifdef _WIN32
  #include <windows.h>
#endif

#include "affine.h"
#include "constants.h"
#include "primitives.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <vector>

using namespace std;

class GeometricObject; // definition is in shapes.h

class Ray {
  public:
    // constructors
    Ray(void);
    Ray(const Point3D& start);
    Ray(const Point3D& start, const Vector3D& direction);
    Ray(const Ray& ray);

    // destructor
    ~Ray(void);

    // methods
    void clear(void);
    void copyObjectsPierced(const Ray& ray);
    void copyFromRay(const Ray& ray);
    static void jitterRaysInCone(vector<Ray>& jitteredRays,
                                 const Ray& centralConeRay,
                                 double coneRadius);

    // data-members
    Point3D          fStart;
    Vector3D         fDirection;
    int              fRecurseLevel;
    int              fRow;
    int              fColumn;
    bool             fIsReflective;
    bool             fIsRefractive;
    int              fNrOfObjectsPierced;
    GeometricObject* fObjectsPierced[kMaxObjectsPierced];
    int              fNrOfConeSamples;
    double           fSampleWeight;
    bool             fFollowRay;
    Color            fColor;
};

class HitInfo {
  public:
    // constructor
    HitInfo(void);

    // destructor
    ~HitInfo(void);

    // methods
    void set(const HitInfo& hitInfo);

    // data-members
    double           fHitTime;
    GeometricObject* fObjectHit;
    int              fSurface;
    bool             fIsEntering;
    Point3D          fHitPoint;
    double           fBarycentricU;
    double           fBarycentricV;
    Vector3D         fHitNormal;
};

class Intersection {
  public:
    // constructor
    Intersection(void);

    // destructor
    ~Intersection(void);

    // methods
    void set(const Intersection& intersection);
    bool grazingAtSurface(void);

    // data-members
    int     fNrOfHits;
    HitInfo fHits[kMaxHits];
};

class Transformations {
  public:
    // methods
    static void transformPoint(Point3D& result,
                               const AffineTransformation& affine,
                               const Point3D& point);
    static void translatePoint(Point3D& result,
                               const AffineTransformation& affine,
                               const Point3D& point);
    static void transformVector(Vector3D& result,
                                const AffineTransformation& affine,
                                const Vector3D& vec);
    static void transformNormal(Vector3D& result,
                                const AffineTransformation& affine,
                                const Vector3D& normal);
    static void transformRay(Ray& result,
                             const AffineTransformation& affine,
                             const Ray& ray);
    static Point3D calcRayPosition(const Point3D& start,
                                   const Vector3D& direction,
                                   double distance);
    static Point3D calcRayPosition(const Ray& ray,
                                   double distance);
    static double dot3D(const Vector3D& a, const Vector3D& b);
    static double dot3D(const Point3D& a, const Vector3D& b);
    static double dot3D(const Point3D& a, const Point3D& b);
};

#endif
