// ------------------------------
// Filename      : rgbapixmap.cpp
// Author        : Sven Maerivoet
// Last modified : 14/04/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifdef _WIN32
  #include <windows.h>
#endif

#include "rgbapixmap.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <iostream>

// ----------
// class mRGB
// ----------

// ------------
// constructors
// ------------
mRGBA::mRGBA(void) : fRed(0),
                     fGreen(0),
                     fBlue(0),
                     fAlpha(0)
{
}

mRGBA::mRGBA(const mRGBA& mrgba) : fRed(mrgba.fRed),
                                   fGreen(mrgba.fGreen),
                                   fBlue(mrgba.fBlue),
                                   fAlpha(mrgba.fAlpha)
{
}

mRGBA::mRGBA(uchar red, uchar green, uchar blue) : fRed(red),
                                                   fGreen(green),
                                                   fBlue(blue),
                                                   fAlpha(0)
{
}

mRGBA::mRGBA(uchar red, uchar green, uchar blue, uchar alpha) : fRed(red),
                                                                fGreen(green),
                                                                fBlue(blue),
                                                                fAlpha(alpha)
{
}

// ----------
// destructor
// ----------
mRGBA::~mRGBA(void)
{
}

//--------
// methods
// -------
void mRGBA::set(uchar red, uchar green, uchar blue)
{
  fRed = red;
  fGreen = green;
  fBlue = blue;
}

void mRGBA::set(uchar red, uchar green, uchar blue, uchar alpha)
{
  fRed = red;
  fGreen = green;
  fBlue = blue;
  fAlpha = alpha;
}

void mRGBA::convert(double& red, double& green, double& blue) const
{
  red = fRed / 255.0;
  green = fGreen / 255.0;
  blue = fBlue / 255.0;
}

double mRGBA::normalizeAlpha(void) const
{
  return (((double) fAlpha) / 255.0);
}

// ----------------
// class RGBApixmap
// ----------------

// ------------
// constructors
// ------------
RGBApixmap::RGBApixmap(void) : fNRows(0),
                               fNCols(0),
                               fPixmap(0),
                               fTextureID(0)
{
}

RGBApixmap::RGBApixmap(int rows, int cols) : fNRows(rows),
                                             fNCols(cols),
                                             fTextureID(0)
{
  // allocate memory to store pixelmap
  fPixmap = new mRGBA[fNRows * fNCols];
}

// ----------
// destructor
// ----------
RGBApixmap::~RGBApixmap(void)
{
  clear();
}

// -------
// methods
// -------
void RGBApixmap::clear(void)
{
  fNRows = 0;
  fNCols = 0;
  delete[] fPixmap;
  fPixmap = 0;
}

bool RGBApixmap::loadFromBMPFile(string filename)
{
  fFile.open(filename.c_str(),ios::in|ios::binary);
  if (!fFile) {
    cerr << "RGBApixmap::loadFromBMPFile() -> can't open file [" << filename << "]." << endl;
    return false;
  }

  // read the header
  char letterB;
  fFile.get(letterB);
  char letterM;
  fFile.get(letterM);
  ulong fileSize       = readLong();
  ushort reserved1     = readShort();
  ushort reserved2     = readShort();
  ulong  offBits       = readLong();
  ulong  headerSize    = readLong();
  ulong  numCols       = readLong();
  ulong  numRows       = readLong();
  ushort planes        = readShort();
  ushort bitsPerPixel  = readShort();
  ulong  compression   = readLong();
  ulong  imageSize     = readLong();
  ulong  xPels         = readLong();
  ulong  yPels         = readLong();
  ulong  numLUTentries = readLong();
  ulong  impColors     = readLong();

  // check the pixel-depth (true-color)
  if (bitsPerPixel != 24) {
    fFile.close();
    cerr << "RGBApixmap::loadFromBMPFile() -> incorrect #bits per pixel [" << bitsPerPixel
         << "] in file [" << filename << "], should be 24." << endl;
    return false;
  }

  // round up (3 * numCols) to next multiple of 4
  int nBytesInRow = (((3 * numCols) + 3) / 4) * 4;
  int numPadBytes = nBytesInRow - (3 * numCols);

  fNRows = numRows;
  fNCols = numCols;
  fPixmap = new mRGBA[fNRows * fNCols];
  if (fPixmap == 0) { // memory allocation failed
    cerr << "RGBApixmap::loadFromBMPFile() -> memory allocation failed." << endl;
    fFile.close();
    return false;
  }

  long count = 0;

  cerr << "Loading [" << filename << "] ";
  int oneNinth = fNRows / 9;
  int rowsDone = 0;
  int blocksDone = 0;

  for (int row = 0; row < fNRows; ++row) {

    ++rowsDone;
    if (rowsDone >= oneNinth) {
      rowsDone = 0;
      if ((blocksDone % 2) == 0) {
        cerr << "-";
      }
      else {
        cerr << "=";
      }
      ++blocksDone;
    }

    for (int col = 0; col < fNCols; ++col) {
      char red, green, blue;
      fFile.get(blue);
      fFile.get(green);
      fFile.get(red);
      fPixmap[count].fRed = red;
      fPixmap[count].fGreen = green;
      fPixmap[count].fBlue = blue;
      fPixmap[count].fAlpha = 0;
      ++count;
    }

    // skip the padded bytes at each row's end
    for (int k = 0; k < numPadBytes; ++k) {
      char dummy;
      fFile >> dummy;
    }
  }

  cerr << " done." << endl;

  fFile.close();
  return true;
}

bool RGBApixmap::writeToBMPFile(string filename)
{
  fFile.open(filename.c_str(),ios::out|ios::binary);
  if (!fFile) {
    cerr << "RGBApixmap::writeToBMPFile() -> can't open file [" << filename << "]." << endl;
    return false;
  }

  // round up (3 * numCols) to next multiple of 4
  ushort nBytesInRow = ((3 * fNCols + 3) / 4) * 4;
  int numPadBytes = nBytesInRow - (3 * fNCols);
  ulong biSizeImage = nBytesInRow * ((ulong) fNRows);

  // write the header
  ushort bfType = 0x4d42; // 'BM';'B'=0x42, 'M'=0x4d
  writeShort(bfType);
  ulong bfSize = 54 + biSizeImage; // total size of image = header-size + biSizeImage
  writeLong(bfSize);
  writeShort(0); // reserved 1 
  writeShort(0); // reserved 2
  writeLong(ulong(54)); // bfOffBits : 54 - but not used in readers
  writeLong(ulong(40)); // biSize - bytes in info header
  writeLong(ulong(fNCols));
  writeLong(ulong(fNRows));
  writeShort(ushort(1));
  writeShort(ushort(24)); // bit planes & bit count
  writeLong(0L); // compression
  writeLong(biSizeImage);
  writeLong(0L); // pelsPerMeterX
  writeLong(0L); // pelsPerMeterY
  writeLong(0L); // colors used
  writeLong(0); // important colors

  long count = 0;

  for (int row = 0; row < fNRows; ++row) {
    
    for (int col = 0; col < fNCols; ++col) {
      fFile.put(fPixmap[count].fBlue);
      fFile.put(fPixmap[count].fGreen);
      fFile.put(fPixmap[count].fRed);
      // discard the alpha-channel
      ++count;
    }

    // now pad this row so num bytes is a mult of 4
    for (int k = 0; k < numPadBytes; ++k) {
      fFile.put(char(0));
    }
  }

  fFile.close();
  return true;
}

void RGBApixmap::setTextureID(int textureID)
{
  fTextureID = textureID;
}

int RGBApixmap::getTextureID(void) const
{
  return fTextureID;
}

int RGBApixmap::getNrOfRows(void) const
{
  return fNRows;
}

int RGBApixmap::getNrOfColumns(void) const
{
  return fNCols;
}

void RGBApixmap::setRGBATriplet(int u, int v, const mRGBA& mrgba)
{
  fPixmap[(v * fNCols) + u] = mrgba;
}

mRGBA RGBApixmap::getRGBATriplet(int u, int v) const
{
  return fPixmap[(v * fNCols) + u];
}

void RGBApixmap::getColor(int u, int v, double& red, double& green, double& blue) const
{
  mRGBA mrgba = getRGBATriplet(u,v);
  mrgba.convert(red,green,blue);
}

mRGBA* RGBApixmap::getPixmap(void) const
{
  return fPixmap;
}

bool RGBApixmap::copyFromScreen(int left, int bottom,
                                int width, int height)
{
  fNRows = height;
  fNCols = width;

  fPixmap = new mRGBA[fNRows * fNCols];
  if (fPixmap == 0) { // memory allocation failed
    cerr << "RGBApixmap::copyFromScreen() -> memory allocation failed." << endl;
    return false;
  }

  // tell OpenGL NOT to try to align pixels to 4 byte boundaries in memory
  glPixelStorei(GL_PACK_ALIGNMENT,1);
  glReadPixels(left,bottom,fNCols,fNRows,GL_RGBA,GL_UNSIGNED_BYTE,fPixmap);

  return true;
}

void RGBApixmap::drawToScreen(int left, int bottom) const
{
  if ((fNRows == 0) ||
      (fNCols == 0)) {
    return;
  }

  // set the current raster-position
  glRasterPos2i(left,bottom);

  // tell OpenGL NOT to try to align pixels to 4 byte boundaries in memory
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  glDrawPixels(fNCols,fNRows,GL_RGBA,GL_UNSIGNED_BYTE,fPixmap);
}

void RGBApixmap::copyFromScreenToScreen(int left, int bottom,
                                        int width, int height,
                                        int newLeft, int newBottom) const
{
  // set the current raster-position
  glRasterPos2i(newLeft,newBottom);

  glCopyPixels(left,bottom,width,height,GL_COLOR);
}

void RGBApixmap::setChromaKey(const mRGBA& chromaKey)
{
  long count = 0;

  for (int row = 0; row < fNRows; ++row) {
    for (int col = 0; col < fNCols; ++col) {

      mRGBA p = fPixmap[count];

      if ((p.fRed == chromaKey.fRed) &&
          (p.fGreen == chromaKey.fGreen) &&
          (p.fBlue == chromaKey.fBlue)) {
        // make transparant
        fPixmap[count].fAlpha = 0;
      }
      else {
        // make opaque
        fPixmap[count].fAlpha = 255;
      }

      ++count;
    }
  }
}

void RGBApixmap::blendOnScreen(int left, int bottom) const
{
  // result = (alpha * source) + ((1 - alpha) * destination)
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  drawToScreen(left,bottom);
  glDisable(GL_BLEND);
}

// ---------------
// utility methods
// ---------------
ushort RGBApixmap::readShort(void)
{
  // the BMP-format uses little-endian integer types
  char ic;
  ushort ip;

  fFile.get(ic);
  ip = ic;

  fFile.get(ic);
  ip |= ((ushort) ic << 8);

  return ip;
}

ulong RGBApixmap::readLong(void)
{
  // the BMP-format uses little-endian integer types
  char ic = 0;
  ulong ip = 0;

  uchar uc = ic;
  fFile.get(ic);
  uc = ic;
  ip = uc;

  fFile.get(ic);
  uc = ic;
  ip |= ((ulong) ic << 8);

  fFile.get(ic);
  uc = ic;
  ip |= ((ulong) ic << 16);

  fFile.get(ic);
  uc = ic;
  ip |= ((ulong) ic << 24);

  return ip;
}

void RGBApixmap::writeShort(ushort us)
{
  // the BMP-format uses little-endian integer types
  fFile.put((char) (us & 0xff));
  fFile.put((char) ((us >> 8) & 0xff));
}

void RGBApixmap::writeLong(ulong ul)
{
  // write long in little-endian form
  fFile.put((char) (ul & 0xff));
  fFile.put((char) ((ul >> 8) & 0xff));
  fFile.put((char) ((ul >> 16) & 0xff));
  fFile.put((char) ((ul >> 24) & 0xff));
}
