// ------------------------------
// Filename      : rng.cpp
// Author        : Sven Maerivoet
// Last modified : 05/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#include "rng.h"
#include <assert.h>
#include <math.h>
#include <time.h>

/*
  This is a multiply-with-carry type of random number generator
  invented by George Marsaglia.  The algorithm is :

    S = 2111111111 * X[n-4] + 1492 * X[n-3] + 1776 * X[n-2] + 5115 * X[n-1] + C
    X[n] = S modulo 2^32
    C = floor(S / 2^32)

  This implementation uses a long double for C. Note that not all C++
  compilers support the long double (80-bit) floating point format.
  You will get an error message if your compiler doesn't support this.
*/

// ---------
// class RNG
// ---------

// initialization of static class-members
double RNG::fX[5] = {0.0,0.0,0.0,0.0,0.0};

// -----------
// constructor
// -----------
RNG::RNG(void)
{
}

// ----------
// destructor
// ----------
RNG::~RNG(void)
{
}

// -------
// methods
// -------
void RNG::setSeed(uint32 seed)
{
  // check that the compiler supports 80-bit long doubles
  assert(sizeof(long double) > 6);

  // make sure seed != 0
  if (seed == 0) {
    seed = (uint32) -1;
  }

  int i;

  // make first 5 random numbers and put them into the buffer
  for (i = 0; i < 5; ++i) {
    seed ^= seed << 13;
    seed ^= seed >> 17;
    seed ^= seed << 5;
    fX[i] = seed * (1.0 / (65536.0 * 65536.0));
  }

  // randomize some more
  for (i = 0; i < 20; ++i) {
    double dummy = uniform();
  }
}

void RNG::setSeedWithTime(void)
{
  setSeed(time(0));
}

int RNG::uniformInt(void)
{
  return ((int) (32767.0 * uniform()));
}

double RNG::uniform(void)
{
  long double c = (2111111111.0 * fX[3]) +
                  (1492.0 * (fX[3] = fX[2])) +
                  (1776.0 * (fX[2] = fX[1])) +
                  (5115.0 * (fX[1] = fX[0])) +
                  fX[4];
  fX[4] = floorl(c);
  fX[0] = c - fX[4];
  fX[4] = fX[4] * (1.0 / (65536.0 * 65536.0));

  return fX[0];
}
