// ------------------------------
// Filename      : shapes.h
// Author        : Sven Maerivoet
// Last modified : 16/06/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

/*
  The class-tree for this file is as follows :
  --------------------------------------------

  |
  +-- Material
  |
  +-- GeometricObject
      |
      +-- Boolean
      |   |
      |   +-- UnionBool
      |   +-- IntersectionBool
      |   +-- DifferenceBool
      |
      +-- Shape
          |
          +-- Cube
          +-- Mesh
          +-- Sphere
          +-- TaperedCylinder
          +-- Torus
          |
          +-- AlgebraicSurface
              |
              |   // (quadric surfaces)
              +-- OneSheetHyperboloid
              +-- TwoSheetHyperboloid
              +-- EllipticCylinder
              +-- EllipticCone
              +-- EllipticParaboloid
              +-- HyperbolicParaboloid
              |
              |   // (cubic surfaces)
              +-- PinchedSurfaceWithHoles
              +-- PinchedSurfaceWithLobe
              |
              |   // (quartic surfaces)
              +-- FourLobes
              +-- SteinersRomanSurface
              +-- ToothCube
*/

#ifndef SM_SHAPES_H // include-guard
#define SM_SHAPES_H

#ifdef _WIN32
  #include <windows.h>
#endif

#include "affine.h"
#include "constants.h"
#include "extents.h"
#include "primitives.h"
#include "proceduraltexturer.h"
#include "ray.h"
#include "rgbapixmap.h"
#include "solidtexturer.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <string>

using namespace std;

class Material {
  public:
    // constructor
    Material(void);

    // destructor
    ~Material(void);

    // methods
    void setDefault(void);
    void set(const Material& material);

    // data-members
    typedef SolidTexturer::ESolidTextureType SolidTextureType;
    typedef ProceduralTexturer::EProceduralTextureType ProceduralTextureType;
    double                fSurfaceRoughness;
    Color                 fEmissive;
    Color                 fAmbient;
    Color                 fDiffuse;
    Color                 fSpecular;
    double                fSpecularExponent;
    bool                  fLightBackFaces;
    double                fReflectivity;
    double                fGlossStrength;
    double                fTransparency;
    int                   fPriority;
    double                fTranslucencyStrength;
    double                fSpeedOfLight;
    bool                  fRetainAmbientColor;
    bool                  fRetainDiffuseColor;
    bool                  fRetainSpecularReflection;
    bool                  fDisableRefraction;
    SolidTextureType      fSolidTextureType;
    double                fSolidTextureParameters[kMaxSolidTextureParameters];
    bool                  fHasImageTexture;
    RGBApixmap*           fImageTextureBitmap;
    double                fImageTextureTileS;
    double                fImageTextureTileT;
    ProceduralTextureType fProceduralTextureType;
    double                fProceduralTextureParameters[kMaxProceduralTextureParameters];
    bool                  fFlatTextureIsModulating;
    double                fFlatTextureSOffset;
    double                fFlatTextureTOffset;
    bool                  fMotionBlurEnabled;
};

class GeometricObject {
  public:
    // constructor
    GeometricObject(void);

    // destructor
    virtual ~GeometricObject(void);

    // methods
    virtual void computeExtent(void);
    void drawExtent(void) const;
    void setMaterial(const Material& material);
    virtual void drawOpenGL(void) const;
    virtual void tellMaterialsGL(void) const;
    virtual bool computeIntersectionsWithRay(const Ray& ray, Intersection& intersection);
    virtual void textureLookup(const HitInfo& hitInfo, double& s, double& t) const;
    void computeTextureCoordinates(const HitInfo& hitInfo, double& s, double& t) const;
    virtual void applyMotionBlurTransformation(double timeIndex, double frameLength);
    virtual void restoreOriginalTransformation(void);

    // data-members
    GenericExtent* fWorldExtent;
    Material       fMaterial;

  protected:
    // data-members
    bool fWarningShown;
};

class Boolean : public GeometricObject {
  public:
    // constructor
    Boolean(void);

    // destructor
    virtual ~Boolean(void);

    // methods
    virtual void drawOpenGL(void) const;
    virtual bool computeIntersectionsWithRay(const Ray &ray, Intersection& intersection);
    void applyMotionBlurTransformation(double timeIndex, double frameLength);
    void restoreOriginalTransformation(void);

    // data-members
    GeometricObject* fLeft;
    GeometricObject* fRight;
};

class UnionBool : public Boolean {
  public:
    // constructor
    UnionBool(void);

    // destructor
    ~UnionBool(void);

    // methods
    bool computeIntersectionsWithRay(const Ray &ray, Intersection& intersection);
};

class IntersectionBool : public Boolean {
  public:
    // constructor
    IntersectionBool(void);

    // destructor
    ~IntersectionBool(void);

    // methods
    bool computeIntersectionsWithRay(const Ray &ray, Intersection& intersection);
};

class DifferenceBool : public Boolean {
  public:
    // constructor
    DifferenceBool(void);

    // destructor
    ~DifferenceBool(void);

    // methods
    bool computeIntersectionsWithRay(const Ray &ray, Intersection& intersection);
};

class Shape : public GeometricObject {
  public:
    // constructor
    Shape(void);

    // destructor
    virtual ~Shape(void);

    // methods
    virtual void computeExtent(void);
    virtual void drawOpenGL(void) const;
    void tellMaterialsGL(void) const;
    virtual bool computeIntersectionsWithRay(const Ray& ray, Intersection& intersection);
    virtual void textureLookup(const HitInfo& hitInfo, double& s, double& t) const;
    void applyMotionBlurTransformation(double timeIndex, double frameLength);
    void restoreOriginalTransformation(void);

    // data-members
    AffineTransformation fAffineTransformation;
    AffineTransformation fInverseAffineTransformation;
    AffineTransformation fOriginalAffineTransformation;
    AffineTransformation fOriginalInverseAffineTransformation;
    double               fMotionBlurRotationAngle;
    Vector3D             fMotionBlurRotationAxis;
    Vector3D             fMotionBlurTranslation;
    Vector3D             fMotionBlurScale;
};

class Cube : public Shape {
  public:
    // constructor
    Cube(void);

    // destructor
    ~Cube(void);

    // methods
    void drawOpenGL(void) const;
    bool computeIntersectionsWithRay(const Ray& ray, Intersection& intersection);
    void textureLookup(const HitInfo& hitInfo, double& s, double& t) const;

  private:
    // utility methods
    Vector3D calcNormal(int surface) const;
};

class VertexID {
  public:
    // data-members
    int fVertexIndex;
    int fNormalIndex;
};

class Face {
  public:
    // constructor
    Face(void);

    // destructor
    ~Face(void);

    // data-members
    int       fNrOfVertices;
    VertexID* fVertices;
};

class Mesh : public Shape {
  public:
    // the different filetypes
    enum EMeshType {k3VN,
                    k3DV,
                    k3DM,
                    kOBJ};

    // constructor
    Mesh(string filename, EMeshType meshType);

    // destructor
    ~Mesh(void);

    // methods
    void readFromFile(string filename, EMeshType meshType);
    void writeToFile(string filename, EMeshType meshType) const;
    bool isEmpty(void) const;
    void clear(void);
    void triangulateFaces(void);
    bool isTriangulated(void) const;
    void computeExtent(void);
    void drawOpenGL(void) const;
    bool computeIntersectionsWithRay(const Ray& ray, Intersection& intersection);
    void textureLookup(const HitInfo& hitInfo, double& s, double& t) const;

    // data-members
    bool fInterpolateNormals;

  private:
    // utility methods
    bool isCounterClockWise(const Point2D& p1, const Point2D& p2, const Point2D& p3) const;
    double lengthSquared(const Point2D& p1, const Point2D& p2) const;

    // data-members
    int       fNrOfVertices;
    int       fNrOfNormals;
    int       fNrOfFaces;
    int       fNrOfTriangulatedFaces;
    Point3D*  fVertices;
    Vector3D* fNormals;
    Face*     fFaces;
    Face*     fTriangulatedFaces;
    int       fLastVertexUsed;
    int       fLastNormalUsed;
    int       fLastFaceUsed;
    string    fFilename;
    bool      fIsTriangulated;
};

class Sphere : public Shape {
  public:
    // constructor
    Sphere(void);

    // destructor
    ~Sphere(void);

    // methods
    void drawOpenGL(void) const;
    bool computeIntersectionsWithRay(const Ray& ray, Intersection& intersection);
    void textureLookup(const HitInfo& hitInfo, double& s, double& t) const;
};

class TaperedCylinder : public Shape {
  public:
    // constructor
    TaperedCylinder(void);

    // destructor
    ~TaperedCylinder(void);

    // methods
    void drawOpenGL(void) const;
    bool computeIntersectionsWithRay(const Ray& ray, Intersection& intersection);
    void textureLookup(const HitInfo& hitInfo, double& s, double& t) const;

    // data-members
    double fSmallRadius;
    bool   fDistortTexture;
};

class Torus : public Shape {
  public:
    // constructor
    Torus(void);

    // destructor
    ~Torus(void);

    // methods
    void computeExtent(void);
    void drawOpenGL(void) const;
    bool computeIntersectionsWithRay(const Ray& ray, Intersection& intersection);
    void textureLookup(const HitInfo& hitInfo, double& s, double& t) const;

    // data-members
    double fTubeRadius, fTorusRadius;

  private:
    // utlitity functions
    Vector3D calcNormal(const Point3D& p) const;
};

class AlgebraicSurface : public Shape {
  public:
    // constructor
    AlgebraicSurface(void);

    // destructor
    virtual ~AlgebraicSurface(void);

    // methods
    void computeExtent(void);
    void drawOpenGL(void) const;
    bool computeIntersectionsWithRay(const Ray& ray, Intersection& intersection);
    void textureLookup(const HitInfo& hitInfo, double& s, double& t) const;

    // data-members
    double fBoxExtentWidth, fBoxExtentHeight, fBoxExtentDepth;

  protected:
    // algebraic surface specific functions
    virtual void setCoefficients(const Ray& genericRay);
    virtual Vector3D calcNormal(const Point3D& p) const;

    // data-members
    double fCoefficients[kMaxAlgebraicSurfaceDegree + 1];
};

class OneSheetHyperboloid : public AlgebraicSurface {
  public:
    // constructor
    OneSheetHyperboloid(void);

    // destructor
    ~OneSheetHyperboloid(void);

  protected:
    // quadratic surface specific methods
    void setCoefficients(const Ray& genericRay);
    Vector3D calcNormal(const Point3D& p) const;
};

class TwoSheetHyperboloid : public AlgebraicSurface {
  public:
    // constructor
    TwoSheetHyperboloid(void);

    // destructor
    ~TwoSheetHyperboloid(void);

  protected:
    // quadratic surface specific methods
    void setCoefficients(const Ray& genericRay);
    Vector3D calcNormal(const Point3D& p) const;
};

class EllipticCylinder : public AlgebraicSurface {
  public:
    // constructor
    EllipticCylinder(void);

    // destructor
    ~EllipticCylinder(void);

  protected:
    // quadratic surface specific methods
    void setCoefficients(const Ray& genericRay);
    Vector3D calcNormal(const Point3D& p) const;
};

class EllipticCone : public AlgebraicSurface {
  public:
    // constructor
    EllipticCone(void);

    // destructor
    ~EllipticCone(void);

  protected:
    // quadratic surface specific methods
    void setCoefficients(const Ray& genericRay);
    Vector3D calcNormal(const Point3D& p) const;
};

class EllipticParaboloid : public AlgebraicSurface {
  public:
    // constructor
    EllipticParaboloid(void);

    // destructor
    ~EllipticParaboloid(void);

  protected:
    // quadratic surface specific methods
    void setCoefficients(const Ray& genericRay);
    Vector3D calcNormal(const Point3D& p) const;
};

class HyperbolicParaboloid : public AlgebraicSurface {
  public:
    // constructor
    HyperbolicParaboloid(void);

    // destructor
    ~HyperbolicParaboloid(void);

  protected:
    // quadratic surface specific methods
    void setCoefficients(const Ray& genericRay);
    Vector3D calcNormal(const Point3D& p) const;
};

class PinchedSurfaceWithHoles : public AlgebraicSurface {
  public:
    // constructor
    PinchedSurfaceWithHoles(void);

    // destructor
    ~PinchedSurfaceWithHoles(void);

  protected:
    // cubic surface specific methods
    void setCoefficients(const Ray& genericRay);
    Vector3D calcNormal(const Point3D& p) const;
};

class PinchedSurfaceWithLobe : public AlgebraicSurface {
  public:
    // constructor
    PinchedSurfaceWithLobe(void);

    // destructor
    ~PinchedSurfaceWithLobe(void);

  protected:
    // cubic surface specific methods
    void setCoefficients(const Ray& genericRay);
    Vector3D calcNormal(const Point3D& p) const;
};

class FourLobes : public AlgebraicSurface {
  public:
    // constructor
    FourLobes(void);

    // destructor
    ~FourLobes(void);

  protected:
    // quartic surface specific methods
    void setCoefficients(const Ray& genericRay);
    Vector3D calcNormal(const Point3D& p) const;
};

class SteinersRomanSurface : public AlgebraicSurface {
  public:
    // constructor
    SteinersRomanSurface(void);

    // destructor
    ~SteinersRomanSurface(void);

  protected:
    // quartic surface specific methods
    void setCoefficients(const Ray& genericRay);
    Vector3D calcNormal(const Point3D& p) const;
};

class ToothCube : public AlgebraicSurface {
  public:
    // constructor
    ToothCube(void);

    // destructor
    ~ToothCube(void);

  protected:
    // quartic surface specific methods
    void setCoefficients(const Ray& genericRay);
    Vector3D calcNormal(const Point3D& p) const;
};

#endif
