// ------------------------------
// Filename      : sdlviewer.cpp
// Author        : Sven Maerivoet
// Last modified : 30/05/2001
// Target        : C++ (OpenGL)
//
// All rights reserved.
// ------------------------------

#ifdef _WIN32
  #include <windows.h>
#endif

#include "sdlviewer.h"
#include "opengl/opengl.h"
#include "opengl/primitives.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <string.h>

#include <iostream>

using namespace std;

// --------------------
// class SDLViewerGLApp
// --------------------

// -----------
// constructor
// -----------
SDLViewerGLApp::SDLViewerGLApp(string filename) : fFilename(filename)
{
}

// ----------
// destructor
// ----------
SDLViewerGLApp::~SDLViewerGLApp(void)
{
}

// -----------------------
// explicit initialisation
// -----------------------
void SDLViewerGLApp::explicitInitialize()
{
  fEnableRayTracing = false;
  fBlockSize = 8;
  fNrOfConeSamples = 10;
  fComputeShadows = true;
  fShadowComputationMode = kHardShadows;
  fAntiAlias = false;
  fAntiAliasMode = kUse5Samples;
  fUseStochasticSampling = false;
  fNormalizeColors = false;
  fNormalizeBrightness = 1.1;
  fShowCoordinates = false;
  fFindNormals = false;
  fFollowRays = false;

  setCamera(Point3D(0.0,2.0,3.0),
            Point3D(0.0,0.0,0.0),
            Vector3D(0.0,1.0,0.0));
  setCameraPerspectiveShape(45.0,getWorldWindowAspectRatio(),
                            0.1,1000.0);

  setCameraSlideStepSize(0.1);

  loadScene(fFilename);
  setSceneViewSettings();
}

// ------------------
// callback-functions
// ------------------
void SDLViewerGLApp::displayFunc(void)
{
  enableSceneLighting();
  switchLightingOn();
  drawScene();

  if (fEnableRayTracing) {

    EShadowComputationMode shadowComputationMode(fShadowComputationMode);
    if (!fComputeShadows) {
      shadowComputationMode = kNoShadows;
    }

    EAntiAliasMode antiAliasMode(fAntiAliasMode);
    if (!fAntiAlias) {
      antiAliasMode = kUse1Sample;
    }

    rayTraceScene(fBlockSize,
                  fNrOfConeSamples,
                  shadowComputationMode,
                  0.05,
                  antiAliasMode,
                  fUseStochasticSampling,
                  0.1,
                  fNormalizeColors,
                  fNormalizeBrightness,
                  fFindNormals,
                  0.001,
                  fFollowRays,
                  0.001);
  }

  if (fShowCoordinates) {
    Point3D eyePoint = getCameraEyePoint();
    Point3D viewPoint = getCameraViewPoint();

    drawString(10,15,"eye = (" +
                     convertDoubleToString(eyePoint.fX) + "," +
                     convertDoubleToString(eyePoint.fY) + "," +
                     convertDoubleToString(eyePoint.fZ) + ")",
               kYellow);
  }
}

void SDLViewerGLApp::keyboardFunc(unsigned char key, int x, int y)
{
  TOpenGLApp::keyboardFunc(key,x,y);

  switch (key) {
    case 'r' :
      fEnableRayTracing = false;
      enableSceneLighting();
      break;

    case 'R' :
      fEnableRayTracing = true;
      break;

    case '-' :
      fBlockSize /= 2;
      if (fBlockSize < 1) {
        fBlockSize = 1;
      }
      break;

    case '+' :
      fBlockSize *= 2;
      break;

    case 'd' :
      fComputeShadows = false;
      break;

    case 'D' :
      fComputeShadows = true;
      break;

    case 'h' :
      fShadowComputationMode = kHardShadows;
      break;

    case 'H' :
      fShadowComputationMode = kSoftShadows;
      break;

    case 'u' :
      fAntiAlias = false;
      break;

    case 'U' :
      fAntiAlias = true;
      break;

    case 'b' :
      fAntiAliasMode = kUse5Samples;
      fNrOfConeSamples = 10;
      break;

    case 'B' :
      fAntiAliasMode = kUse9Samples;
      fNrOfConeSamples = 30;
      break;

    case 'g' :
      fUseStochasticSampling = false;
      break;

    case 'G' :
      fUseStochasticSampling = true;
      break;

    case 'i' :
      fNormalizeColors = false;
      enableSceneLighting();
      break;

    case 'I' :
      fNormalizeColors = true;
      break;

    case 'p' :
      fNormalizeBrightness -= 0.1;
      if (fNormalizeBrightness < 0.1) {
        fNormalizeBrightness = 0.1;
      }
      break;

    case 'P' :
      fNormalizeBrightness += 0.1;
      break;

    case 'k' :
      fShowCoordinates = false;
      break;

    case 'K' :
      fShowCoordinates = true;
      break;

    case 'w' :
      fFindNormals = false;
      break;

    case 'W' :
      fFindNormals = true;
      break;

    case 'j' :
      fFollowRays = false;
      break;

    case 'J' :
      fFollowRays = true;
      break;
  }

  if (fScheduleRedisplay) {
    redisplay();
  }
}
